#pragma once

#include <iostream>
#include <math.h>
#include <stdio.h>

#include <vector>
#include <stdlib.h>

#include <armadillo>

#include "General_Functions.h"

using namespace std;
using namespace arma;

mat Vij_x_CS_function (
	const unsigned short &TrainDim, 
	const unsigned short &DimIn, 
	const mat &TrainSet, 
	const double &sig2)
{
	mat Vij_x(TrainDim,TrainDim);
		
	// ##### Calculus Vij_inputs #####
	//mat Vij_x(TrainDim,TrainDim);
	rowvec dij(DimIn); // Vi_x(TrainDim);
	double aux_vix = 0.0, aux_vx= 0.0 /*, V_x = 0.0*/;

	for (unsigned short i=0; i<TrainDim; i++)
	{
		for (unsigned short j=0; j<TrainDim; j++)
		{
			for (unsigned short in=0; in<DimIn; in++)
			{
				dij(in) = TrainSet(i,in) - TrainSet(j,in);
			}
			//dij = TrainSet.row(i) - TrainSet.row(j);
			Vij_x(i,j) = Vij_function(dij, DimIn, sig2);	
		}
	}
	// ##### End Calculus Vij_inputs #####
	return (Vij_x);
}



void CS_Weights1Bias1_function (
	const unsigned short &DimHidden, 
	const unsigned short &DimIn, 
	const unsigned short &TrainDim, 
	const mat &HiddenTrain, 
	const mat &TrainSet, 
	const mat &Vij_x, 
	const colvec &Vi_x,
	const float &actFcnFlatness, 
	const double &sig2, 
	const float &momRate, 
	const float &lrCoeff1, 
	mat &Weights1, //esta matriz  alterada directamente na actual
	const mat &prev_Weights1, 
	rowvec &Bias1, //este vector  alterado directamente no actual
	const rowvec &prevBias1, 
	const ActivationFunctions &NeuronActFunct)
{
	
	double aux1 = 0.0;
	double dTrainDim = static_cast<double>(TrainDim);

	// ##### Calculus of Vij concerning the hidden layer #####
	mat Vij_y(TrainDim,TrainDim);
	rowvec dij_y(DimHidden), Vi_y(TrainDim);
	double aux_viy = 0.0;

	for (unsigned short i=0; i<TrainDim; i++)
	{
		for (unsigned short j=0; j<TrainDim; j++)
		{
			for (unsigned short hid=0; hid<DimHidden; hid++)
			{
				dij_y(hid) = HiddenTrain(i,hid) - HiddenTrain(j,hid);
			}
			Vij_y(i,j) = Vij_function(dij_y, DimHidden, sig2);
			aux_viy += Vij_y(i,j);
		}
		Vi_y(i) = aux_viy/dTrainDim;
		aux_viy = 0.0;
	}
	// ##### End of Calculus of Vij concerning the hidden layer #####

	// ##### Auxiliares no Calculo Parcela 1 e 2: D1 e D2 #####
	double aux_vj = 0.0, aux_vm = 0.0;
	for (unsigned short i=0; i<TrainDim; i++)
	{
		for (unsigned short j=0; j<TrainDim; j++)
		{
			aux_vj += Vij_x(i,j) * Vij_y(i,j);
			aux_vm += Vij_y(i,j);
		}
	}
	// ##### Fim Auxiliares no Calculo Parcela 1 e 2: D1 e D2 #####

	//##### BackForward Step == MaxMI_Cauchy-Swartz Criterium #####
	vec y_i(DimHidden), y_j(DimHidden), Fi(DimHidden);
	vec dij(DimHidden), auxN1(DimHidden), auxN2(DimHidden), auxN3(DimHidden), N1(DimHidden), N2(DimHidden), N3(DimHidden);
	
	double D1=0.0, D2=0.0, D3=0.0;
	D1 = aux_vj;
	D2 = aux_vm;
	N1.zeros(), N2.zeros(), N3.zeros();
													

	for (unsigned short i=0; i<TrainDim; i++) //Cross samples: sample t
	{
		for (unsigned short j=0; j<TrainDim; j++) //Cross samples: sample s
		{
			for(unsigned short hid=0; hid<DimHidden; hid++) //for each output neuron
			{
				y_i(hid) = HiddenTrain(i,hid);
				y_j(hid) = HiddenTrain(j,hid);
				dij(hid) = y_i(hid) - y_j(hid);

				// ############ dMI/dyi = Fi
				// ver Principe Tutorial pag.25 Eq.28
				auxN1(hid) = Vij_x(i,j) * Vij_y(i,j) * dij(hid);
				auxN2(hid)= Vij_y(i,j) * dij(hid);
				auxN3(hid) = Vi_x(j) * Vij_y(i,j) * dij(hid);
			}
			N1 += auxN1;
			N2 += auxN2;
			N3 += auxN3;
			D3 += Vi_x(j) * Vi_y(j);
		}

		// ############ dMI/Dw = dMI/dy * dy/dw ############				
		mat delta_weights1(DimIn,DimHidden);
		for(unsigned short hid=0; hid<DimHidden; hid++) //for each output neuron: derivative
		{
			Fi(hid) = -1.0/( dTrainDim*dTrainDim * sig2 ) * ( N1(hid)/D1 + N2(hid)/D2 - 2.0*N3(hid)/D3 );

			//Derivative of ACTIVATION FUNCTION @ Y_s
			double derv_i = DervActFcn(y_i(hid), actFcnFlatness, NeuronActFunct);

			//dMI/dyi = Fi
			//dy/dw = derv_ActFn_t * x_t - derv_ActFn_s * x_s
			for (unsigned short in=0; in<DimIn; in++)
			{
				double momterm = momRate * ( Weights1(in,hid) - prev_Weights1(in,hid)); //momentum term
					
				delta_weights1(in,hid) = ( (1.0) * lrCoeff1 * Fi(hid) * (derv_i * TrainSet(i,in) ) ) + momterm; //MaxMI = Max (-log IP) = Min log IP
				Weights1(in,hid) += delta_weights1(in,hid);
			}
									
			Bias1(hid) += ( (1.0) * lrCoeff1 * Fi(hid) * derv_i ) + ( momRate * ( Bias1(hid) - prevBias1(hid) ) );
		}
	}
	
}


void CostFunction_CS (
	double &CostFn,  //para ser atualizada diretamente
	const unsigned short &DimHidden, 
	const unsigned short &NumPatterns, 
	const mat &HiddenMatrix, 
	const double &sig2, 
	const mat &Vij_x, 
	const colvec &Vi_x, 
	const double &V_x)
{
	//########## Calculating Cost Function for this Output ##########
	double dNumPatterns = static_cast<double>(NumPatterns);

	double aux_CostFn = 0.0, aux_Vi_CF_y = 0.0, aux_V_CF_y = 0.0, V_CF_y = 0.0;
	mat Vij_CF_y(NumPatterns,NumPatterns);
	rowvec Vi_CF_y(NumPatterns);

	rowvec dij_CF_y(DimHidden);
	for (unsigned short i=0; i<NumPatterns; i++)
	{
		for (unsigned short j=0; j<NumPatterns; j++)
		{
			for (unsigned short hid=0; hid<DimHidden; hid++)
			{
				dij_CF_y(hid) = HiddenMatrix(i,hid) - HiddenMatrix(j,hid);
			}
			Vij_CF_y(i,j) = Vij_function(dij_CF_y, DimHidden, sig2);

			aux_Vi_CF_y += Vij_CF_y(i,j);
		}
		Vi_CF_y(i) = aux_Vi_CF_y / dNumPatterns;
		aux_V_CF_y += Vi_CF_y(i);
	}
	V_CF_y = aux_V_CF_y / dNumPatterns;

	double aux_VJ_CF = 0.0, aux_VC_CF = 0.0;
	double VJ_CF=0.0, VM_CF=0.0, VC_CF=0.0; //the 3 parts of the cost function: joint, marginal and cross information potentials
		 
	for (unsigned short i=0; i<NumPatterns; i++)
	{
		for (unsigned short j=0; j<NumPatterns; j++)
		{
			aux_VJ_CF += Vij_x(i,j) * Vij_CF_y(i,j);
		}
		aux_VC_CF += Vi_x(i) * Vi_CF_y(i);
	}
	VJ_CF = aux_VJ_CF / ( dNumPatterns*dNumPatterns ); //the argument for the joint information potential
	aux_VJ_CF = 0.0;
	VM_CF = V_x * V_CF_y; //the argument for the marginal information potential
	VC_CF = aux_VC_CF / ( dNumPatterns ); //the argunment for the cross information potential
	aux_VC_CF = 0.0;
							
	CostFn = (log(VJ_CF) + log(VM_CF) - 2.0*log(VC_CF)); //the complete information potential

}