#pragma once

#include <iostream>
#include <math.h>
#include <stdio.h>

#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <stdlib.h>


#include <armadillo> //matrices

#include <boost/random.hpp>

using namespace std;
using namespace arma;


#define W1_file
//#define W2_file
#define B1_file
//#define B2_file

//mt importante, tem o MSE do train, test e output, finais (ao fim de todas as epocas)
// o melhor para cada seed, de entre todos os c1 e c2 testados
#ifndef MSE_Validation_file
#define MSE_Validation_file
#endif

//resultados ao fim de todas as epocas
#ifndef ResultsFile
//#define ResultsFile
#endif

//resultados para cada epoca (MSE)
#ifndef DetailsFile
//#define DetailsFile
#endif

//retorna para cada epoca o valor da cost function da primeira parte (so p itl methods)
#ifndef CostFnTestFile
#define CostFnTestFile
#endif

//ficheiro com o tempo em segundos para todos os c1 e c2, so para o treino da rede
#ifndef TimeFile
#define TimeFile
#endif

//sai as duas matrizes de validation a pura e a 100%, para o estudo com denoising ou sem (puro), caso jakov
#ifndef ValidationOutputFiles
#define ValidationOutputFiles
#endif
//sai um ficheiro com o mse, a seed e a matriz de outputs dos target 
//#define JeanValidationOutputFiles


enum GLOBAL_Train 
{
	ITL,
	PROP
};


//Activation Functions
enum ActivationFunctions 
{
	NEURON_ACT_LOGISTIC,
	NEURON_ACT_TANH,
	NEURON_ACT_LINEAR
};

enum sigmaOption 
{
	SILVERMAN_RULE,
	PRINCIPE_RULE
};

enum OjaRule 
{
	OjaYES,
	OjaNO
};


enum TrainMethods_W1 
{
	ENTROPY,
	CAUCHY_SCHWARTZ,
	EUCLIDEAN_DISTANCE,
	UNI
};

enum TrainMethods_W2 
{
	W2_PROP,
	CIM,
	Corr
};

enum InitializationW1Methods 
{
	RANDOM_W1,
	PCA
};

enum InitializationW2Methods 
{
	RANDOM_W2,
	TRANSPOSE
};

enum InitializationB1Methods 
{
	RANDOM_B1,
	ZEROS_B1
};

enum InitializationB2Methods 
{
	RANDOM_B2,
	ZEROS_B2
};


enum NormalizationMethod
{
	ENTRANCE_MIN_MAX,
	GLOBAL_MIN_MAX,
	ENTRANCE_ZSCORE,
	GLOBAL_ZSCORE,
	NONE
};



//definio da funo de aleatorios entre -1 e 1
boost::random::uniform_real_distribution<> aleat(-1,1);


double gaussKernel(
	const double &e, 
	const double &sig2, 
	const unsigned short &m) //Helper function for MULTI kernel calculations
{
	double pi = 3.14159265359;
	double dm = static_cast<double>(m);

	double gk = (1.0/( sqrt( pow(2.0*pi,dm) * pow(sig2,dm) ))) * exp(-0.5*((e)/sig2));
	
	return (gk);
}



double Vij_function(const rowvec &dij, const unsigned short &Dim, const double &sig2) // entropy
{
		//Given dij, Dim of vec dij and sig2
		//Vij = G(dij,2*sig2)
		float aux=0.0f;
		
		//Calculating V_ij
		for (unsigned short k=0; k < Dim; k++)
		{
			aux += (dij(k)*dij(k));
		}
		double a =2.0*sig2;
		double V_ij = gaussKernel(aux, a, Dim);
		return (V_ij);
}



double  actFcn(const double &activation, const float &actFcnFlatness, const ActivationFunctions &NeuronActFunct) 
{
	double output;

	switch (NeuronActFunct)
	{
		case NEURON_ACT_LOGISTIC: 
		{
			output = 1 / ( 1 + exp( - activation / actFcnFlatness ) );
			break;
		}
		
		case NEURON_ACT_TANH:
		{
			// output = tanh( activation / actFcnFlatness );
			output = ( 2.0 / ( 1.0 + exp( - activation / actFcnFlatness ) ) ) - 1.0; //KEKO
			//
			// double u = activation / actFcnFlatness;
			// output = (exp(u) - exp(-u))/(exp(u)+exp(-u));
			break;
		}
		
		case NEURON_ACT_LINEAR:
		{
			if(activation<-1)
			{
				output=-1;
			}
			else if(activation>1)
			{
				output=1;
			}
			else 
			{
				output = activation;
			}
			break;
		}

		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
		
	return (output);
}

double  DervActFcn(
	double output, //tem de ficar copia
	const float &actFcnFlatness, 
	const ActivationFunctions &NeuronActFunct) 
{
	double tmpDerivative;

	switch (NeuronActFunct)
	{
		case NEURON_ACT_LOGISTIC: 
		{
			tmpDerivative = ( output * ( 1 - output ) ) / actFcnFlatness;
			break;
		}
		
		case NEURON_ACT_TANH:
		{
			tmpDerivative = ( 1.0 - (output*output) ) / ( 2.0 * actFcnFlatness ); 
			break;
		}
		
		case NEURON_ACT_LINEAR:
		{
			tmpDerivative = 1;  
			break;
		}

		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}

	return (tmpDerivative);
}


//functions of initialization of weights
void PCA_function(
	mat &Weights1, //para ser actualizada
	const mat &TrainSet_PCA, 
	const unsigned short &DimIn, 
	const unsigned short &DimHidden)
{
	mat Coeff_PCA(DimIn,DimIn);
	Coeff_PCA = princomp(TrainSet_PCA);
	Weights1 = Coeff_PCA.submat( span(), span(0, DimHidden-1) );
}

//read files
mat lerFicheiro	(
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const string &Filename)
{
	//##### Reading Test Data from csv file #######
	ifstream indata1;
	double f1;
	string s1="";
		
	indata1.open(Filename);
	if(!indata1) { cerr << "Error: file could not be opened" << endl; exit(1);}

	vector <double> auxtest;
	while(getline(indata1, s1, ';'))
	{
		stringstream fs( s1 );
		fs >> f1;
		auxtest.push_back(f1);
	}

	mat aux_TestSet(linhas,colunas); // test set matrix
	for(unsigned short te=0; te<linhas; te++)
	{
		for (unsigned short in=0; in<colunas; in++)
		{
			aux_TestSet(te,in) = auxtest[(te*colunas+in)];
		}
	}
	indata1.close();
	return (aux_TestSet);
	//##### END of Reading Test Data from csv file #######
}




void transitionLayer (
	mat &AfterMatrix, //para ser actualizado
	const unsigned short &NumberPatterns, 
	const unsigned short &DimFirstLayer, 
	const unsigned short &DimSecondLayer, 
	const mat &originalMatrix, 
	const mat &Weights, 
	const rowvec &Bias, 
	const float &actFcnFlatness, 
	const ActivationFunctions &NeuronActFunct)
{
	double aux1 = 0.0;
	
	for (unsigned short tr=0; tr<NumberPatterns; tr++)
	{
		for(unsigned short hid=0; hid<DimSecondLayer;hid++)
		{
			for(unsigned short in=0; in<DimFirstLayer;in++)	
			{
				aux1 += ((originalMatrix(tr,in))*Weights(in,hid));
			}
			aux1 += Bias(hid);
			AfterMatrix(tr,hid) = actFcn(aux1, actFcnFlatness, NeuronActFunct);
			aux1 = 0.0;
		}
	}
}


mat nnOutputsFunction (
	const unsigned short &NumberPatterns, 
	const unsigned short &DimIn, 
	const unsigned short &DimHidden, 
	const unsigned short &DimOut, 
	const mat &Weights1, 
	const mat &Weights2, 
	const rowvec &Bias1, 
	const rowvec &Bias2, 
	const float &actFcnFlatness, 
	const mat &InMatrix, 
	const ActivationFunctions &NeuronActFunct1, 
	const ActivationFunctions &NeuronActFunct2)
{
	//compresso
	mat HiddenMatrix (NumberPatterns, DimHidden);
	transitionLayer (HiddenMatrix, NumberPatterns, DimIn, DimHidden, InMatrix, Weights1, Bias1, actFcnFlatness, NeuronActFunct1);

	//descompresso
	mat OutputMatrix (NumberPatterns, DimOut);
	transitionLayer (OutputMatrix, NumberPatterns, DimHidden, DimOut, HiddenMatrix, Weights2, Bias2, actFcnFlatness, NeuronActFunct2);

	return (OutputMatrix);
}


//error function

double MAEfunction (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &ErrorMatrix)
{
	double MAE = 0.0;
	for(unsigned short te=0; te<linhas; te++)
	{
		for(unsigned short out=0; out<colunas; out++)
		{
			MAE += abs(ErrorMatrix(te,out));
		}
	}
	double dLinhas = static_cast<double>(linhas);
	double dColunas = static_cast<double>(colunas);
	MAE = MAE/(dLinhas*dColunas);
	return (MAE);
}

double MSEfunction (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &ErrorMatrix)
{
	double MSE = 0.0;
	for(unsigned short tr=0; tr<linhas; tr++)
	{
		for(unsigned short out=0; out<colunas; out++)
		{
			MSE += (ErrorMatrix(tr,out)*ErrorMatrix(tr,out));
		}
	}

	double dLinhas = static_cast<double>(linhas);
	double dColunas = static_cast<double>(colunas);

	MSE = MSE/(dLinhas*dColunas);
	return (MSE);
}



double sigmaFunction (
	const unsigned short &DimHidden, 
	const unsigned short &TrainDim, 
	const mat &HiddenTrain, 
	const sigmaOption &sigmaMethod,
	const float &AuxSigma)
{
	//##### Sigma Calculation #####
	//commun calculations

	double dTrainDim = static_cast<double>(TrainDim);
	double dDimHidden = static_cast<double>(DimHidden);

	vec meane (DimHidden); 
	vec meane2 (DimHidden);
	meane.zeros(); 
	meane2.zeros();
	for (unsigned short tr=0; tr<TrainDim; tr++)
	{
		for (unsigned short hid=0; hid<DimHidden; hid++)
		{
			meane(hid) += HiddenTrain(tr,hid)/dTrainDim;
			meane2(hid) += HiddenTrain(tr,hid)*HiddenTrain(tr,hid);
		}
	}
	double aux2 = 0.0;
	for (unsigned short hid=0; hid<DimHidden; hid++)
	{
		aux2 += meane2(hid)/(dTrainDim-1.0) - meane(hid)*meane(hid)* dTrainDim/(dTrainDim-1.0); //sum of marginals variance
	}

	double sdev = aux2 / (dDimHidden);
	double sig2;

	switch (sigmaMethod)
	{
		case SILVERMAN_RULE: 
		{
			// Sigma Calculation By Silverman's Rule
			// Silverman, 1998 (First Edition 1986), page 87
			//in case of a NN with a single neuron in the hidden layer, this rule would lead to the simplification
			//simplification: sig = 1.06 * sample standard deviation * N^(-1/5) - 1DimIn
			//the general case, that is for any number of hidden neurons (one or more than one neuron), the expression is as follows:
			double Ak = pow((4.0/(dDimHidden+2.0)), 1.0/(dDimHidden+4.0));
			double h_opt = Ak * pow(static_cast<double> (dTrainDim), -1.0/(dDimHidden+4.0));
			double sig = sqrt(sdev) * h_opt;
			sig2 = sig*sig*AuxSigma;
			break;
		}
		case PRINCIPE_RULE: 
		{
			// clculo segundo o PRINCIPE ITL (book) page58 eq2.15
			double Ak = pow(4.0/(2.0*dDimHidden+1.0), 1.0/(dDimHidden+4.0));
			double h_opt = Ak * pow((dTrainDim), -1.0/(dDimHidden+4.0));
			double sig = sqrt(sdev) * h_opt;
			sig2 = sig*sig*AuxSigma;
			break;
		}
		
		default: cout << "ERROR!!! ";
	}
	return (sig2);
}

float maxLrCoeffFunction (
	const unsigned short &DimIn, 
	const mat &TrainSet, 
	const double &learningInicial)
{
	//##### Maximum Learning Coeficient Calculation #####
	
	// CRITERIUM: lrCoeff < 1/ 2*lambdaMax
	//ver livro do Hagan, pg. 9-6
	// cov = 1/(n-1) sum( (xi-xm)(xj-xm) ) 
	
	
	//COM BASE NA COVARINCIA
	mat C1(DimIn,DimIn); //, C2(DimIn,DimIn);
	C1 = cov(TrainSet,TrainSet);
	vec eigval1 = eig_sym(C1);
	double lambda1 = max(eigval1);
	float Max_lrCoeff1 = 1.0f / ( 2.0f*static_cast<float>(lambda1));
	
	
	//COM BASE NA CORRELAO 
	/*
	mat dd = stddev (TrainSet);
	unsigned short ExisteZero=0;
	double Max_lrCoeff1;

	for (unsigned short i=0; i<DimIn; i++)
	{
		if(dd(0,i)==0)
		{
			ExisteZero=1;
		}
	}

	if (ExisteZero==1)
	{
		//faz o max learning rate constante
		Max_lrCoeff1 = learningInicial;
	}
	else
	{
		//calcula a correlaao e o maxlearning rate
		mat C1(DimIn,DimIn); 
		C1 = cor(TrainSet,TrainSet);
		vec eigval1 = eig_sym(C1);
		double lambda1 = max(eigval1);
		Max_lrCoeff1 = 1.0 / (lambda1);
	}
	*/
	
	//cout << "lrCoeff1 = " << lrCoeff1 << endl;
	//cout << "Maximum lrCoeff1 = " << Max_lrCoeff1 << endl;
	//##### End of Maximum Learning Coeficient Calculation  #####

	return (Max_lrCoeff1);
}


float SatLevelFunction (
	const unsigned short &TrainDim, 
	const unsigned short &colunas, 
	const mat &Matrix, 
	const float &SatLimit)
{
	mat aux_saturation(TrainDim,colunas); //auxiliary matrix to calculate the proportion of weigths within saturation state (the weights the absoulte value higher than 0.9) 
	aux_saturation.zeros();

	for (unsigned short tr=0; tr<TrainDim; tr++)
	{
		for(unsigned short out=0; out<colunas; out++)
		{


			if (abs(Matrix(tr,out))>=SatLimit) //o limite de saturao est definido em cima em conjunto com os outros parametros
			{
				// the matrix will the value one for the corresponding positions of saturated weights (for an absolute specified value, e.g. 0.9)
				aux_saturation(tr,out) =1;
			}
			else
			{
				aux_saturation(tr,out) =0;
			}

		}
	}

	double dTrainDim = static_cast<double>(TrainDim);
	double dcolunas = static_cast<double>(colunas);

	double aux_SatLevel=accu(aux_saturation); //retorna a soma acumulada de todos os elementos que esto contidos dentro da matrix auxiliar
	float SatLevel=static_cast<float>(aux_SatLevel)/(static_cast<float>(dTrainDim)*static_cast<float>(dcolunas)); //retorna a percentagem de weights 2 que esto saturados
	
	return (SatLevel);
}

void Saturation_Process (
	const unsigned short &TrainDim, 
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &TrainSet,
	const float &SatLimit,
	const float &ThresholdSat,
	const float &actFcnFlatness, 
	const ActivationFunctions &NeuronActFunct1,
	mat &Evaluated_Matrix, 
	mat &Weights1,
	rowvec &Bias1,
	boost::random::mt19937 &rng
	)
{
	float SatLevel = SatLevelFunction (TrainDim, colunas, Evaluated_Matrix, SatLimit);
					
	if(SatLevel>ThresholdSat)
	{
		for(unsigned short hid=0; hid<colunas;hid++)
		{
			for (unsigned short in=0; in<linhas; in++) 
			{
				Weights1(in,hid) =  0.1*aleat(rng)*Weights1(in,hid)+Weights1(in,hid);	
			}
			Bias1(hid) = 0.1*aleat(rng)*Bias1(hid)+Bias1(hid);
		}
		transitionLayer (Evaluated_Matrix, TrainDim, linhas, colunas, TrainSet, Weights1, Bias1, actFcnFlatness, NeuronActFunct1);					
	}
}


mat ErrorMatrixFunction (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &Matrix1, 
	const mat &Matrix2)
{
	//Error matrix calculation

	mat ErrorMatrix (linhas, colunas);

	for(unsigned short tr=0; tr<linhas; tr++)
	{
		for (unsigned short out=0; out < colunas; out++)
		{
			ErrorMatrix(tr,out) = Matrix1(tr,out) - Matrix2(tr,out);
		}
	}
	return (ErrorMatrix);
}

double Error_MSE_function (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &Matrix1, 
	const mat &Matrix2)
{
	mat ErrorMatrix = ErrorMatrixFunction (linhas, colunas, Matrix1, Matrix2);
	double MSE = MSEfunction (linhas, colunas, ErrorMatrix);
	return (MSE);
}


string doubletoStringFunction (double a)
{
	ostringstream learn2;
	learn2 << a;
	string str2 = learn2.str();
	return (str2);
}


string StringConstructor_function (
	const double &c1, 
	const string &MethodName, 
	const string &tipo)
{
	string detailed (MethodName);
	detailed.append("_");
	detailed.append(tipo);
	detailed.append( doubletoStringFunction (c1));
	detailed.append(".csv");
	return (detailed);
}


string StringTrainMethods (const TrainMethods_W1 &Method) 
{
	string MethodName ("");
	switch (Method)
	{
		case ENTROPY: 
			MethodName.append("Entropy"); 
			break;
		
		case CAUCHY_SCHWARTZ: 
			MethodName.append("CauchySchwartz"); 
			break;
			
		case EUCLIDEAN_DISTANCE: 
			MethodName.append("EuclideanDistance"); 
			break;

		case UNI: 
			MethodName.append("Uni"); 
			break;
		
		default: cout << "ERROR!!! ";
	}
	return (MethodName);
}


void learningAdaptCoeff	(
	const double &CostFn, 
	const double &Prev_CostFn, 
	float &lrCoeff, //esta variavel  actualizada directamente	
	const float &lr_inc, 
	const float &Max_lrCoeff, 
	const float &max_perf_inc, 
	const float &lr_dec, 
	const int &maxORmin	)
{
	if(maxORmin==1)
	{
		if( CostFn/Prev_CostFn > 1.0)
		{
			lrCoeff =  lrCoeff * lr_inc; //aceita pesos e aumenta learning rate
			if(lrCoeff > Max_lrCoeff)
			{
				lrCoeff = Max_lrCoeff;
			}
		}
		if( CostFn/Prev_CostFn < (1.0/max_perf_inc)) 
		{
				lrCoeff =  lrCoeff * lr_dec; //corta o learning rate
				if (lrCoeff < 0.00000001f)
				{
					lrCoeff = 0.00000001f;
				}
		}
	}
	else //est a minimizar
	{
		if( CostFn/Prev_CostFn < 1.0)
		{
			lrCoeff =  lrCoeff * lr_inc; //aceita pesos e aumenta learning rate
			if(lrCoeff > Max_lrCoeff)
			{
				lrCoeff = Max_lrCoeff;
			}
		}

		if( CostFn/Prev_CostFn > static_cast<double>(max_perf_inc)) 
		{
			lrCoeff =  lrCoeff * lr_dec; //corta o learning rate
			
			if (lrCoeff < 0.00000001f)
			{
				lrCoeff = 0.00000001f;
			}
		}
	}
	//return (lrCoeff);
}


void learningAdaptMatrix(
	const double &CostFn, 
	const double &Prev_CostFn, 
	const float &max_perf_inc, 
	mat &Matrix, //vai ser alterada a propria matriz de pesos
	const mat &Prev_Matrix, 
	const int &maxORmin)
{
	//##### Learning Coeff Adaptive #####
	if (maxORmin==1)//est a maximizar
	{
		if( CostFn/Prev_CostFn < (1.0/max_perf_inc)) 
		{
			Matrix = Prev_Matrix; //no aceita as alteraes desta poca
		}
	}
	else //est a minimizar
	{
		if( CostFn/Prev_CostFn > max_perf_inc) 
		{	
			Matrix = Prev_Matrix; //no aceita as alteraes desta poca
		}
	}
	//return (Matrix);
}

void learningAdaptBias(
	const double &CostFn, 
	const double &Prev_CostFn, 
	const float &max_perf_inc, 
	rowvec &Bias, //actualiza o BIAS
	const rowvec &Prev_Bias, 
	const int &maxORmin)
{
	//##### Learning Coeff Adaptive #####
	if (maxORmin==1)//est a maximizar
	{
		if( CostFn/Prev_CostFn < (1.0/max_perf_inc)) 
		{
			Bias = Prev_Bias;
		}
	}
	else //est a minimizar
	{
		if( CostFn/Prev_CostFn > max_perf_inc) 
		{	
			Bias = Prev_Bias;
		}
	}
	//##### End Learning Coeff Adaptative #####
}


//funo de estimar o cim para matrizes, no  usado mas pode ser util um dia
double CIM_Estimator (
	const unsigned short &DimOut, 
	const unsigned short &NumPatterns, 
	const mat &OutMatrixS,
	const mat &OutMatrixO, 
	const double &sig2
	)
{
	double aux_Correntropy = 0.0;
	double dNumPatterns = static_cast<double>(NumPatterns);
	double dij3=0.0; double zero=0.0;
	unsigned short um=1;
	for (unsigned short i=0; i<NumPatterns; i++) //usa-se indice i e j por percorrer as combinaes dois a dois
	{
			for (unsigned short out=0; out<DimOut; out++)
			{
				dij3 = OutMatrixO(i,out) - OutMatrixS(i,out);
				double bb=dij3*dij3;
				aux_Correntropy += gaussKernel(bb, sig2, um);
			}
	}
	
	double Correntropy = ( 1.0 / (dNumPatterns*DimOut) ) * aux_Correntropy;
	
	double KernelZero = gaussKernel(zero, sig2, um);
	//ver paper Liu, Pokharel and Principe 2007 - CIM
	double CIM = pow(KernelZero - Correntropy, 0.5);
	
	return (CIM);
}


double Corr_Estimator_fromVectors (
	const vec &vetor1, 
	const vec &vetor2, 
	const unsigned short &DimOut, 
	const double &sig2)
{
	//nota: a correntropia  simtrica, a ordem dos vectores no  relevante
	double aux_Correntropy = 0.0;
	double dij3;
	unsigned short um = 1;
	for (unsigned short out=0; out<DimOut; out++)
	{
		dij3 = vetor1(out) - vetor2(out);
		double bb= dij3*dij3;
		aux_Correntropy += gaussKernel(bb, sig2, um);
	}
	
	double Correntropy = ( 1.0 / (static_cast<double>(DimOut)) ) * aux_Correntropy;
	return (Correntropy);
}


double CIM_Estimator_fromVectors (
	const unsigned short &DimOut, 
	const vec &OutvecS,
	const vec &OutvecO, 
	const double &sig2
	)
{
	double zero =0.0;
	unsigned short um = 1;
	
	double Correntropy = Corr_Estimator_fromVectors (OutvecS, OutvecO, DimOut, sig2);
	
	double KernelZero = gaussKernel(zero, sig2, um);
	
	//ver paper Liu, Pokharel and Principe 2007 - CIM
	double CIM = pow(KernelZero - Correntropy, 0.5);
	
	return (CIM);
}




//CASO W1B1: Dim2=DimHidden, Dim1=dimIn; CASO W2B2: Dim2=dimOut, Dim1=DimHidden
void OjaFunction (
	const unsigned short &Dim2, 
	const unsigned short &Dim1, 
	mat& Weights, 
	rowvec& Bias)
{	
	//clculo da regra de OJA
	vec denominadorOJA (Dim2);
	double auxOja=0.0;
					
	for (unsigned short hid=0; hid<Dim2; hid++)
	{
		auxOja=0.0;
		for (unsigned short in=0; in<Dim1; in++)
		{
			auxOja += Weights(in,hid) * Weights(in,hid);
		}
		//inclui o valor do bias
		auxOja += Bias(hid) * Bias(hid);
		denominadorOJA (hid)= sqrt(auxOja);
	}

	//actualizao dos pesos
	for (unsigned short hid=0; hid<Dim2; hid++)
	{
		for (unsigned short in=0; in<Dim1; in++)
		{						
			Weights(in, hid) = Weights(in,hid) / denominadorOJA (hid);
		}
		Bias(hid) = Bias(hid)/ denominadorOJA (hid);
	}
}