//Autoencoder that applies the maximization of Entropy in hidden layer to train the first half of the network
//and minimizes the error between the target and the input layers to train the second half of the autoencoder (classical PROP with adaptive learning rates)
//the ITL definition is applied in a multivariate approach (covariance matrix with diagonal only, same sigma for all positions)
//code developed  @ INESC Porto by Vera Palma and Joana Martins - December 25 (ye, its Xmas!) 2012

using namespace std;

#if 0

#include <iostream>
#include <math.h>
#include <stdio.h>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <stdlib.h>
#include <algorithm> //sort

//chama o ficheiro function network
#include "NetworkFunction.h"
#include "Normalization_functions.h"
#include "Prop_Network.h"
#include "ProblemSpecifications2AA.h"
#include "MoreFunctions2AAonly.h"



//armadillo libraries
#include <armadillo> //matrices
using namespace arma;


// ------------- NN MaxE Implementation -------------
int main()
{
	
	Specifications_function_2AA (
	learnCoeff1, learnCoeff2, numEpochs1, numEpochs2, W1_Method, W2_Method, Train_Global, ChosenBreaker, TestDim_A, ValidationDim_A, TrainDim_A, TestDim_F, ValidationDim_F, TrainDim_F,
	DimIn, DimHidden, DimOut, TrainSetFile_A, TestSetFile_A, ValidationSetFile_A, TargetTrSetFile_A, TargetTsSetFile_A, TargetVlSetFile_A, Label_data_A, TrainSetFile_F,
	TestSetFile_F, ValidationSetFile_F, TargetTrSetFile_F, TargetTsSetFile_F, TargetVlSetFile_F, Label_data_F, ValidationSet_B, ValidationKey_B);
		
	mat notNormTrainSet_A = lerFicheiro	(TrainDim_A, DimIn, TrainSetFile_A); // Reading Training Data from csv file 
	mat notNormTrainSet_F = lerFicheiro	(TrainDim_F, DimIn, TrainSetFile_F); // Reading Training Data from csv file 

	// ##### Data normalization considering "Aberto" and "Fechado" simultaneously #####
	//define the vectores with maximum and minimum values of the train set
	
	rowvec Parameters1_A = Parameter1_function (notNormTrainSet_A, NormMethod, DimIn); //extract the vector with maximum values of TRAIN set
	rowvec Parameters1_F = Parameter1_function (notNormTrainSet_F, NormMethod, DimIn); 
	rowvec Parameters2_A = Parameter2_function (notNormTrainSet_A, NormMethod, DimIn);
	rowvec Parameters2_F = Parameter2_function (notNormTrainSet_F, NormMethod, DimIn);
	

	mat TrainSet_A = NormalizationTrain_function (Parameters1_A,  Parameters2_A,  notNormTrainSet_A,  TrainDim_A,  DimIn,  NormMethod); // normalize TRAIN set (using max and min of TRAIN set)
	mat TrainSet_F = NormalizationTrain_function (Parameters1_F,  Parameters2_F,  notNormTrainSet_F,  TrainDim_F,  DimIn,  NormMethod); // normalize TRAIN set (using max and min of TRAIN set)
	
	mat TestSet_A = ReadAndNorm_function (TestDim_A, DimIn, TestSetFile_A, Parameters1_A, Parameters2_A, NormMethod); //read & normalize TEST set (using max and min of TRAIN set)
	mat TestSet_F = ReadAndNorm_function (TestDim_F, DimIn, TestSetFile_F, Parameters1_F, Parameters2_F, NormMethod); //read & normalize TEST set (using max and min of TRAIN set)
	
	//meramente auxiliar
	mat ValidationSet_A = ReadAndNorm_function (ValidationDim_A, DimIn, ValidationSetFile_A, Parameters1_A, Parameters2_A, NormMethod); //read & normalize VALIDATION set (using max and min of TRAIN set)
	mat ValidationSet_F = ReadAndNorm_function (ValidationDim_F, DimIn, ValidationSetFile_F, Parameters1_F, Parameters2_F, NormMethod); //read & normalize VALIDATION set (using max and min of TRAIN set)
	
	//conjunto de validao Final 
	mat ValidationSet_forA   = ReadAndNorm_function (TotalValidationDim, DimIn, ValidationSet_B, Parameters1_A, Parameters2_A, NormMethod);
	mat ValidationSet_forF   = ReadAndNorm_function (TotalValidationDim, DimIn, ValidationSet_B, Parameters1_F, Parameters2_F, NormMethod);
	

	mat TargetTrSet_A=TrainSet_A; 
	mat TargetTrSet_F=TrainSet_F;
	mat TargetTsSet_A=TestSet_A; 
	mat TargetTsSet_F=TestSet_F;
	mat TargetVlSet_A=ValidationSet_A; 
	mat TargetVlSet_F=ValidationSet_F;
	
	double Best_MSE_test_A =100.0; 
	double Best_MSE_test_F =100.0;

	mat ValidationKey = lerFicheiro	(TotalValidationDim, 1, ValidationKey_B);
	
		
	ofstream file;
	
	boost::random::mt19937 rng;

	for(unsigned short seed=StartSeedNumber; seed < (StartSeedNumber+ passo*NumSeeds); seed+=passo)
	{
		mat  W1_A; rowvec B1_A; mat W2_A; rowvec B2_A;
		mat  W1_F; rowvec B1_F; mat W2_F; rowvec B2_F;

		double Best_MSE_test_A;
		double Best_MSE_test_F;
		

		rng.seed(seed);

		
		//REDE ABERTA
		switch (Train_Global)
		{
			case PROP: 
				{
				NetworkPROP_function(
						W1_A, B1_A, W2_A, B2_A, 
						Init_W1_Method, 
						Init_W2_Method,
						Init_B1_Method,
						Init_B2_Method,
						Label_data_A, 
						NeuronActFunct1,
						NeuronActFunct2,
						OjaDecision,
						seed,
						rng,//new
						DimIn, 
						DimHidden, 
						DimOut, 
						TrainDim_A, 
						TestDim_A, 
						ValidationDim_A, 
						learnCoeff1,
						TrainSet_A, 
						TestSet_A, 
						ValidationSet_A, 
						TargetTrSet_A,
						TargetTsSet_A,
						TargetVlSet_A,
						numEpochs1, 
						lr_dec, 
						lr_inc, 
						max_perf_inc, 
						goal,
						momRate, 
						actFcnFlatness, 
						SatLimit, 
						ThresholdSat
					);

					
					break;
				}
			case ITL: 
				{
					NetworkFunction (
						W1_A, B1_A, W2_A, B2_A, 
						W1_Method, 
						W2_Method,
						Init_W1_Method,
						Init_W2_Method,
						Init_B1_Method,
						Init_B2_Method,
						Label_data_A,
						sigmaMethod,
						NeuronActFunct1,
						NeuronActFunct2,
						OjaDecision,
						seed,
						rng,//new 
						DimIn, 
						DimHidden, 
						DimOut, 
						TrainDim_A, 
						TestDim_A, 
						ValidationDim_A,
						learnCoeff1,
						learnCoeff2, 
						TrainSet_A,
						TestSet_A,
						ValidationSet_A,
						TargetTrSet_A,
						TargetTsSet_A,
						TargetVlSet_A,
						numEpochs1,
						numEpochs2,
						lr_dec, //for the learning adaptive decreasing rate (see Hagan book)
						lr_inc, //for the learning adaptide increasing rate (see Hagan book)
						max_perf_inc, //maximum performance increase racio (=CostFunction/Prev_CostFunction) a partir do qual nao se actualizam pesos
						goal, //como no matlab
						momRate, //momentum coef
						actFcnFlatness, // For Activation Function
						SatLimit, //saturation limit 
						ThresholdSat, //saturation threshold
						AuxSigma,
						AuxSigma2,//for sigma_corr
						L,//for sigma_corr
						lrCoeff_sigma//for sigma_corr
						); 
			
					break;
				}
		
			default: cout << "ERROR!!! ";
		}
		
		mat TestOutputMain_A = nnOutputsFunction (TestDim_A, DimIn, DimHidden, DimOut, W1_A, W2_A, B1_A, B2_A, actFcnFlatness, TestSet_A, NeuronActFunct1, NeuronActFunct2);
		double sig2_A_testOut = sigmaFunction (DimOut, TestDim_A, TestOutputMain_A, sigmaMethod, AuxSigma2);
		double CIM_Test_A = CIM_Estimator (DimOut, TestDim_A, TestOutputMain_A, TestSet_A, sig2_A_testOut); 
		
		//REDE FECHADA
		switch (Train_Global)
		{
			case PROP: 
				{
				NetworkPROP_function(
					W1_F, B1_F, W2_F, B2_F, 
					Init_W1_Method, 
					Init_W2_Method,
					Init_B1_Method,
					Init_B2_Method,
					Label_data_F, 
					NeuronActFunct1,
					NeuronActFunct2,
					OjaDecision,
					seed,
					rng,//new
					DimIn, 
					DimHidden, 
					DimOut, 
					TrainDim_F, 
					TestDim_F, 
					ValidationDim_F, 
					learnCoeff1,
					TrainSet_F, 
					TestSet_F, 
					ValidationSet_F, 
					TargetTrSet_F,
					TargetTsSet_F,
					TargetVlSet_F,
					numEpochs1, 
					lr_dec, 
					lr_inc, 
					max_perf_inc, 
					goal,
					momRate, 
					actFcnFlatness, 
					SatLimit, 
					ThresholdSat
					);

					break;
				}
			case ITL: 
				{
				NetworkFunction (
					W1_F, B1_F, W2_F, B2_F, 
					W1_Method, 
					W2_Method,
					Init_W1_Method,
					Init_W2_Method,
					Init_B1_Method,
					Init_B2_Method,
					Label_data_F,
					sigmaMethod,
					NeuronActFunct1,
					NeuronActFunct2,
					OjaDecision,
					seed,
					rng,//new 
					DimIn, 
					DimHidden, 
					DimOut, 
					TrainDim_F, 
					TestDim_F, 
					ValidationDim_F,
					learnCoeff1,
					learnCoeff2, 
					TrainSet_F,
					TestSet_F,
					ValidationSet_F,
					TargetTrSet_F,
					TargetTsSet_F,
					TargetVlSet_F,
					numEpochs1,
					numEpochs2,
					lr_dec, //for the learning adaptive decreasing rate (see Hagan book)
					lr_inc, //for the learning adaptide increasing rate (see Hagan book)
					max_perf_inc, //maximum performance increase racio (=CostFunction/Prev_CostFunction) a partir do qual nao se actualizam pesos
					goal, //como no matlab
					momRate, //momentum coef
					actFcnFlatness, // For Activation Function
					SatLimit, //saturation limit 
					ThresholdSat, //saturation threshold
					AuxSigma,
					AuxSigma2,//for sigma_corr
					L,//for sigma_corr
					lrCoeff_sigma//for sigma_corr
					); 

					break;
				}
		
			default: cout << "ERROR!!! ";
		}
				
			
		mat TestOutputMain_F = nnOutputsFunction (TestDim_F, DimIn, DimHidden, DimOut, W1_F, W2_F, B1_F, B2_F, actFcnFlatness, TestSet_F, NeuronActFunct1, NeuronActFunct2);
		double sig2_F_testOut = sigmaFunction (DimOut, TestDim_F, TestOutputMain_F, sigmaMethod, AuxSigma2);
		double CIM_Test_F = CIM_Estimator (DimOut, TestDim_F, TestOutputMain_F, TestSet_F, sig2_F_testOut); 

				
		colvec QuadErrorA =  AvaliationVector_function (TotalValidationDim, DimIn, DimHidden, DimOut, W1_A, W2_A, 
B1_A, B2_A, actFcnFlatness, ValidationSet_forA, NeuronActFunct1, NeuronActFunct2, sig2_A_testOut, FalhasMethod);
			
		colvec QuadErrorF =  AvaliationVector_function (TotalValidationDim, DimIn, DimHidden, DimOut, W1_F, W2_F, 
B1_F, B2_F, actFcnFlatness, ValidationSet_forF, NeuronActFunct1, NeuronActFunct2, sig2_F_testOut, FalhasMethod);


	
		//COMPARAO DOS ERROS E CALCULO DE NUMERO DE FALHAS

		//QuadErrorF, QuadErrorA, mat ValidationKey
		//comparao dos dois vetores
		unsigned short falhas = CountFalhas_function (TotalValidationDim, QuadErrorA, QuadErrorF, ValidationKey, "falhasNormal.csv", FalhasMethod);

		
		//COMPARAO DOS ERROS E CALCULO DE NUMERO DE FALHAS

				
		//vetorFalhas(seed)=falhas;
		file.open("FileFalhas.csv", ios::app);
		file <<
			";Falhas;"<< falhas <<
			";CIM test aberto;"<< CIM_Test_A << 
			";CIM test fechado;"<< CIM_Test_F << 
			";"<<  endl;
		file.close();

	} //fim do loop das seeds

	return 0;
}

#endif