//Autoencoder that applies the maximization of Entropy in hidden layer to train the first half of the network
//and minimizes the error between the target and the input layers to train the second half of the autoencoder (classical PROP with adaptive learning rates)
//the ITL definition is applied in a multivariate approach (covariance matrix with diagonal only, same sigma for all positions)
//code developed  @ USE - INESC-TEC by Vera Palma and Joana da Hora - December 25 (ye, its Xmas!) 2012

#if 1

#include <iostream>
#include <math.h>
#include <stdio.h>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <stdlib.h>
#include <algorithm> //sort

//chama o ficheiro function network
#include "NetworkFunction.h"
#include "Normalization_functions.h"
#include "Prop_Network.h"
//#include "General_Functions.h"
#include "ProblemSpecifications1AA.h"

using namespace std;

//armadillo libraries
#include <armadillo> //matrices
using namespace arma;



// ------------- NN MaxE Implementation -------------
int main()
{
	

	Specifications_function_1AA (Train_Global, numEpochs1, learnCoeff1, numEpochs2, W1_Method, W2_Method, learnCoeff2);
	// ##### Data normalization #####
	mat notNormTrainSet = lerFicheiro	(TrainDim, DimIn, TrainSetFile); // Reading Training Data from csv file

	//MINMAX: Parameter1 = MaxTrain; Z-SCORE: Parameter1 = MeanTrain;
	rowvec Parameter1 =  Parameter1_function (notNormTrainSet,  NormMethod, DimIn); //extract the vector with maximum values of TRAIN set
	//MINMAX: Parameter2 = MinTrain; Z-SCORE: Parameter2 = SDev;
	rowvec Parameter2 = Parameter2_function (notNormTrainSet,  NormMethod, DimIn); //extract the vector with maximum values of TEST set
	
	const mat TrainSet = NormalizationTrain_function ( Parameter1,  Parameter2,  notNormTrainSet,  TrainDim,  DimIn,  NormMethod); // normalize TRAIN set (using max and min of TRAIN set)
	const mat TestSet = ReadAndNorm_function (TestDim, DimIn, TestSetFile, Parameter1, Parameter2, NormMethod); //read & normalize TEST set (using max and min of TRAIN set)
	const mat ValidationSet = ReadAndNorm_function (ValidationDim, DimIn, ValidationSetFile, Parameter1, Parameter2, NormMethod); //read & normalize VALIDATION set (using max and min of TRAIN set)
	
	const mat TargetTrSet = ReadAndNorm_function	(TrainDim, DimOut, TargetTrSetFile,Parameter1, Parameter2, NormMethod);
	const mat TargetTsSet = ReadAndNorm_function	(TestDim, DimOut, TargetTsSetFile, Parameter1, Parameter2, NormMethod);
	const mat TargetVlSet = ReadAndNorm_function	(ValidationDim, DimOut, TargetVlSetFile, Parameter1, Parameter2, NormMethod);
	//double Best_MSE_test=100.0; //inicializado aqui para ser lido ca fora no main

	//##### End Data normalization #####
	boost::random::mt19937 rng;

	for(unsigned short seed=StartSeedNumber; seed < (StartSeedNumber + passo*NumSeeds); seed+=passo)
	{
		mat  W1; rowvec B1; mat W2; rowvec B2;// double Best_MSE_test_Main;
		rng.seed(seed);
		//caso PROP 
		switch (Train_Global)
		{
			case PROP: 
				NetworkPROP_function(
						 W1, B1,W2, B2, 
						 Init_W1_Method, 
						 Init_W2_Method,
						 Init_B1_Method,
						 Init_B2_Method,
						 Label_data, 
						 NeuronActFunct1,
						 NeuronActFunct2,
						 OjaDecision,
						 seed,
						 rng,//new
						 DimIn, 
						 DimHidden, 
						 DimOut, 
						 TrainDim, 
						 TestDim, 
						 ValidationDim, 
						 learnCoeff1,
						 TrainSet, 
						 TestSet, 
						 ValidationSet, 
						 TargetTrSet,
						 TargetTsSet,
						 TargetVlSet,
						 numEpochs1, 
						 lr_dec, 
						 lr_inc, 
						 max_perf_inc, 
						 goal,
						 momRate, 
						 actFcnFlatness, 
						 SatLimit, 
						 ThresholdSat);
				break;
		
			case ITL: 
				NetworkFunction (
						W1, B1,W2, B2, 
						//Best_MSE_test,
						W1_Method, 
						W2_Method,
						Init_W1_Method,
						Init_W2_Method,
						Init_B1_Method,
						Init_B2_Method,
						Label_data,
						sigmaMethod,
						NeuronActFunct1,
						NeuronActFunct2,
						OjaDecision,
						seed,
						rng,//new 
						DimIn, 
						DimHidden, 
						DimOut, 
						TrainDim, 
						TestDim, 
						ValidationDim,
						learnCoeff1,
						learnCoeff2, 
						TrainSet,
						TestSet,
						ValidationSet,
						TargetTrSet,
						TargetTsSet,
						TargetVlSet,
						numEpochs1,
						numEpochs2,
						lr_dec, //for the learning adaptive decreasing rate (see Hagan book)
						lr_inc, //for the learning adaptide increasing rate (see Hagan book)
						max_perf_inc, //maximum performance increase racio (=CostFunction/Prev_CostFunction) a partir do qual nao se actualizam pesos
						goal, //como no matlab
						momRate, //momentum coef
						actFcnFlatness, // For Activation Function
						SatLimit, //saturation limit 
						ThresholdSat, //saturation threshold
						AuxSigma,
						AuxSigma2,//for sigma_corr
						L,//for sigma_corr
						lrCoeff_sigma//for sigma_corr
						);

				break;
			
		
			default: cout << "ERROR!!! ";
		}

		ofstream fileV;	
		
		#ifdef ValidationOutputFiles
		//Outputs do conjunto Validao passando pela rede seleccionada (para cada seed)
		
		mat SaidasHiddenTrain (TrainDim, DimHidden);
		transitionLayer (SaidasHiddenTrain, TrainDim , DimIn, DimHidden, TrainSet, W1, B1, actFcnFlatness, NeuronActFunct1);
		/*
		
		//mat IValidationOutput_cGE = nnOutputsFunction (ValidationDim, DimIn, DimHidden, DimOut, W1, W2, B1, B2, actFcnFlatness, ValidationSet, NeuronActFunct1, NeuronActFunct2);
		//mat Des_IValidationOutput_cGE = GlobalDESNormalizationMinMax (ValidationDim, DimOut, IValidationOutput_cGE, Parameter1, Parameter2);

		

		
		
		fileV.open(QQ_Label_data, ios::app);
		fileV << "Seed:;" << seed << endl;
		for (unsigned int i=0; i<ValidationDim;i++)
		{
			for (unsigned int j=0; j < DimOut; j++)
			{
				fileV << Des_IValidationOutput_cGE(i,j) << ";";
			}
			fileV << endl;
		}
		fileV << endl;
		fileV.close();

		//
		string ValidationSetFile_sGE =	"validation_sGE_b4.csv";	//2500patternsC     MultiAlternadoValidation  validation_sGE_b4 (escolher o nome do ficheiro) 
		mat ValidationSet_sGE = ReadAndNorm_function (ValidationDim, DimIn, ValidationSetFile_sGE, Parameter1, Parameter2, NormMethod); //read & normalize VALIDATION set (using max and min of TRAIN set)
		mat IValidationOutput_sGE = nnOutputsFunction (ValidationDim, DimIn, DimHidden, DimOut, W1, W2, B1, B2, actFcnFlatness, ValidationSet_sGE, NeuronActFunct1, NeuronActFunct2);
		mat Des_IValidationOutput_sGE = GlobalDESNormalizationMinMax (ValidationDim, DimOut, IValidationOutput_sGE, Parameter1, Parameter2);


		QQ_Label_data = ("_Seed_");
		QQ_Label_data.append(str2);
		QQ_Label_data.append("_validationOutput_sGE.csv");
		*/
		ostringstream seedStr;
		seedStr << seed;
		string str2 = seedStr.str();

		string QQ_Label_data = ("_Seed_");
		QQ_Label_data.append(str2);
		QQ_Label_data.append("_Output_Bessa.csv");

		
		fileV.open(QQ_Label_data, ios::app);
		fileV << "Seed:;" << seed << endl;
		for (unsigned int i=0; i<TrainDim;i++)
		{
			for (unsigned int j=0; j < DimHidden; j++)
			{
				fileV << SaidasHiddenTrain(i,j) << ";";
			}
			fileV << endl;
		}
		fileV << endl;
		fileV.close();

		#endif
	
		#ifdef B1_file;
				string b1f (Label_data);
				b1f.append("B1_Final");
				//b1f.append( doubletoStringFunction (seed));
				b1f.append(".csv");
				fileV.open(b1f, ios::app);
				fileV << "seed;" << seed << endl;
				for (unsigned short hid=0; hid<DimHidden; hid++)
				{
					fileV << B1(hid) << ";" ;
				}
				fileV << endl;
				fileV << endl;
				fileV.close();
		#endif

		#ifdef B2_file;
				string b2f (Label_data);
				b2f.append("B2_Final");
				//b2f.append( doubletoStringFunction (seed));
				b2f.append(".csv");
				fileV.open(b2f, ios::app);
				fileV << "seed;" << seed << endl;
				for (unsigned short out=0; out<DimOut; out++)
				{
					fileV << B2(out) << ";" ;
				}
				fileV << endl;
				fileV << endl;
				fileV.close();
		#endif

		#ifdef W1_file;
				string w1f (Label_data);
				w1f.append("W1_Final");
				//w1f.append( doubletoStringFunction (seed));
				w1f.append(".csv");
				fileV.open(w1f, ios::app);
				fileV << "seed;" << seed << endl;
				for (unsigned short in = 0; in< DimIn; in++)
				{
					for (unsigned short hid=0; hid<DimHidden; hid++)
					{
						fileV << W1(in, hid) << ";" ;
					}
					fileV << endl ;
				}
				fileV << endl;
				fileV.close();
		#endif

		#ifdef W2_file;
				string w2f (Label_data);
				w2f.append("W2_Final");
				//	w2f.append( doubletoStringFunction (seed));
				w2f.append(".csv");
				fileV.open(w2f, ios::app);
				fileV << "seed;" << seed << endl;
				for (unsigned short hid = 0; hid< DimHidden; hid++)
				{
					for (unsigned short in=0; in<DimOut; in++)
					{
						fileV << W2(hid, in) << ";" ;
					}
					fileV << endl ;
				}
				fileV << endl;
				fileV.close();
				//fim de files para ver os W1B1W2B2 finais do tipo aberto
		#endif

	} //fim do loop das seeds

	return 0;
}

#endif