
#pragma once

#include <iostream>
#include <math.h>
#include <stdio.h>

#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <stdlib.h>

#include <armadillo> //matrices

//boost libraries
#include <boost/random.hpp>

using namespace std;
using namespace arma;




mat QuadErrorMatrixFunction (unsigned short linhas, unsigned short colunas, mat Matrix1, mat Matrix2)
{
	//Error matrix calculation

	mat ErrorMatrix (linhas, colunas);

	for(unsigned short tr=0; tr<linhas; tr++)
	{
		for (unsigned short out=0; out < colunas; out++)
		{
			ErrorMatrix(tr,out) = (Matrix1(tr,out) - Matrix2(tr,out))*(Matrix1(tr,out) - Matrix2(tr,out));
		}
	}

	return (ErrorMatrix);
}




colvec QuadErrorvec_function (
	unsigned short TotalValidationDim, 
	unsigned short DimIn, 
	unsigned short DimHidden, 
	unsigned short DimOut, 
	mat Weights1, 
	mat Weights2, 
	rowvec Bias1, 
	rowvec Bias2, 
	float actFcnFlatness, 
	mat ValidationSet, 
	ActivationFunctions NeuronActFunct1, 
	ActivationFunctions NeuronActFunct2)
{
	mat OutputsValidation = nnOutputsFunction (
		TotalValidationDim, 
		DimIn, 
		DimHidden, 
		DimOut, 
		Weights1, 
		Weights2, 
		Bias1, 
		Bias2, 
		actFcnFlatness, 
		ValidationSet, 
		NeuronActFunct1, 
		NeuronActFunct2);

	mat ErrorValidationTotal = QuadErrorMatrixFunction  (
		TotalValidationDim, 
		DimOut, 
		ValidationSet,
		OutputsValidation);

	colvec QuadError = sum(ErrorValidationTotal, 1);

	return (QuadError);
}


double CIM_Estimator_fromVectors_2AA 
	(
	unsigned short DimOut, 
	unsigned short NumPatterns, 
	vec OutvecS,
	vec OutvecO, 
	double sig2
	)
{
	double aux_Correntropy = 0.0;
	double dij3=0.0;
	
	for (unsigned short out=0; out<DimOut; out++)
	{
		dij3 = OutvecO(out) - OutvecS(out);
		aux_Correntropy += gaussKernel(dij3*dij3, sig2, 1);
	}
	
	
	double Correntropy = ( 1.0 / static_cast<double>(DimOut) ) * aux_Correntropy;

	double KernelZero = gaussKernel(0.0, sig2, 1);
	//ver paper Liu, Pokharel and Principe 2007 - CIM
	double CIM = pow(KernelZero - Correntropy, 0.5);
	
	return (CIM);
}


colvec CIMvec_function (
	unsigned short TotalValidationDim, 
	unsigned short DimIn, 
	unsigned short DimHidden, 
	unsigned short DimOut, 
	mat Weights1, 
	mat Weights2, 
	rowvec Bias1, 
	rowvec Bias2, 
	float actFcnFlatness, 
	mat ValidationSet, 
	ActivationFunctions NeuronActFunct1, 
	ActivationFunctions NeuronActFunct2,
	double sig2)
{
	mat OutputsValidation = nnOutputsFunction (
		TotalValidationDim, 
		DimIn, 
		DimHidden, 
		DimOut, 
		Weights1, 
		Weights2, 
		Bias1, 
		Bias2, 
		actFcnFlatness, 
		ValidationSet, 
		NeuronActFunct1, 
		NeuronActFunct2);

	colvec CIMvector (TotalValidationDim);
	for (unsigned short i=0; i<TotalValidationDim; i++)
	{
		vec ValidSetVec = ValidationSet(i,span());
		vec ValidOutVec = OutputsValidation(i,span());
		double CIM_of_this_pattern = CIM_Estimator_fromVectors_2AA (DimOut, TotalValidationDim, ValidOutVec, ValidSetVec, sig2);
		CIMvector(i) = CIM_of_this_pattern;
	}
	return (CIMvector);
}

double correntropyFunction_2AA (vec vetor1, vec vetor2, unsigned short DimOut, double sig2)
{
	//nota: a correntropia  simtrica, a ordem dos vectores no  relevante
	double aux_Correntropy = 0.0;
	double dij3;

	for (unsigned short out=0; out<DimOut; out++)
	{
		dij3 = vetor1(out) - vetor2(out);
		aux_Correntropy += gaussKernel(dij3*dij3, sig2, 1);
	}
	
	double Correntropy = ( 1.0 / (static_cast<double>(DimOut)) ) * aux_Correntropy;
	return (Correntropy);
}


colvec CorrVec_function (
	unsigned short TotalValidationDim, 
	unsigned short DimIn, 
	unsigned short DimHidden, 
	unsigned short DimOut, 
	mat Weights1, 
	mat Weights2, 
	rowvec Bias1, 
	rowvec Bias2, 
	float actFcnFlatness, 
	mat ValidationSet, 
	ActivationFunctions NeuronActFunct1, 
	ActivationFunctions NeuronActFunct2,
	double sig2)
{
	mat OutputsValidation = nnOutputsFunction (
		TotalValidationDim, 
		DimIn, 
		DimHidden, 
		DimOut, 
		Weights1, 
		Weights2, 
		Bias1, 
		Bias2, 
		actFcnFlatness, 
		ValidationSet, 
		NeuronActFunct1, 
		NeuronActFunct2);

	colvec CorrVector (TotalValidationDim);
	for (unsigned short i=0; i<TotalValidationDim; i++)
	{
		vec ValidSetVec = ValidationSet(i,span());
		vec ValidOutVec = OutputsValidation(i,span());
		double Corr_of_this_pattern = correntropyFunction_2AA (ValidSetVec, ValidOutVec, DimOut, sig2);
		CorrVector(i) = Corr_of_this_pattern;
	}
	return (CorrVector);
}



colvec AvaliationVector_function (
	unsigned short TotalValidationDim, 
	unsigned short DimIn, 
	unsigned short DimHidden, 
	unsigned short DimOut, 
	mat Weights1, 
	mat Weights2, 
	rowvec Bias1, 
	rowvec Bias2, 
	float actFcnFlatness, 
	mat ValidationSet, 
	ActivationFunctions NeuronActFunct1, 
	ActivationFunctions NeuronActFunct2,
	double sig2,
	FalhasMethods_list FalhasMethod)
{
	colvec AvaliationVector (TotalValidationDim);
	switch (FalhasMethod)
	{
		case MSE_FALHAS: //est-se a minimizar o MSE 
	

			AvaliationVector = QuadErrorvec_function (TotalValidationDim, DimIn, DimHidden, DimOut, Weights1, 
	Weights2, Bias1, Bias2, actFcnFlatness, ValidationSet,	NeuronActFunct1, NeuronActFunct2);
			break;
		
		case CIM_FALHAS: //est-se a minimizar o CIM  
			AvaliationVector = CIMvec_function (TotalValidationDim, DimIn, DimHidden, DimOut, Weights1, 
	Weights2, Bias1, Bias2, actFcnFlatness, ValidationSet,	NeuronActFunct1, NeuronActFunct2, sig2);
	
			break;
			
			
		case Corr_FALHAS: //est-se a maximizar o Corr  
			AvaliationVector = CorrVec_function (TotalValidationDim, DimIn, DimHidden, DimOut, Weights1, 
	Weights2, Bias1, Bias2, actFcnFlatness, ValidationSet,	NeuronActFunct1, NeuronActFunct2, sig2);
			break;

		default: cout << "ERROR!!! ";
	}
	return (AvaliationVector);
}



unsigned short CountFalhas_function	(
	unsigned short TotalValidationDim, 
	colvec AvaliationVectorA, 
	colvec AvaliationVectorF,
	mat ValidationKey, 
	string falhasfile, 
	FalhasMethods_list FalhasMethod)
{
	unsigned short falhas = 0;
	ofstream file;
	file.open(falhasfile, ios::app);

	switch (FalhasMethod)
	{ 
		case MSE_FALHAS: //est-se a minimizar o MSE    
			for (unsigned short n=0; n<TotalValidationDim; n++)
			{
				short auxN=0;
				if(AvaliationVectorF(n)< AvaliationVectorA(n))
				{
					auxN=1;
					file <<	auxN <<";";
				}
				else
				{
					auxN=-1;
					file <<	auxN <<";";
				}
			
				if (auxN!=ValidationKey(n,0))
				{
					falhas+=1;
				}
			}
			break;
		
		case CIM_FALHAS: //est-se a minimizar o CIM     
			for (unsigned short n=0; n<TotalValidationDim; n++)
			{
				short auxN=0;
				if(AvaliationVectorF(n)< AvaliationVectorA(n))
				{
					auxN=1;
					file <<	auxN <<";";
				}
				else
				{
					auxN=-1;
					file <<	auxN <<";";
				}
			
				if (auxN!=ValidationKey(n,0))
				{
					falhas+=1;
				}
			}
			break;
			
			
		case Corr_FALHAS: //est-se a maximizar o Corr   
			for (unsigned short n=0; n<TotalValidationDim; n++)
			{
				short auxN=0;
				if(AvaliationVectorF(n) > AvaliationVectorA(n)) //MAXIMIZAR;  ao contrario dos outros
				{
					auxN=1;
					file <<	auxN <<";";
				}
				else
				{
					auxN=-1;
					file <<	auxN <<";";
				}
			
				if (auxN!=ValidationKey(n,0))
				{
					falhas+=1;
				}
			}
			break;


		default: cout << "ERROR!!! ";
	}


	file <<	endl;
	file.close();

	return (falhas);
}



