#pragma once

#include <iostream>
#include <math.h>
#include <stdio.h>

#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <stdlib.h>

#include <armadillo> //matrices

//boost libraries
#include <boost/random.hpp>

#include <algorithm> //sort
#include <time.h> //time count


#include "Ent_Functions.h"
#include "ED_Functions.h"
#include "CS_Functions.h"
#include "Uni_Functions.h"
#include "W2_functions.h"
#include "General_Functions.h"


using namespace std;
using namespace arma;


//retorna mat de [DimIn+1] por [(2*DimHidden)+1]
void NetworkFunction (
	mat &W1, // para ser alterada directamente
	rowvec &B1, // para ser alterada directamente
	mat &W2, // para ser alterada directamente
	rowvec &B2, // para ser alterada directamente
	TrainMethods_W1 W1_Method,
	TrainMethods_W2 W2_Method,
	const InitializationW1Methods &Init_W1_Method, 
	const InitializationW2Methods &Init_W2_Method, 
	const InitializationB1Methods &Init_B1_Method,
	const InitializationB2Methods &Init_B2_Method,
	const string &Label_data, 
	const sigmaOption &sigmaMethod, 
	const ActivationFunctions &NeuronActFunct1,
	const ActivationFunctions &NeuronActFunct2,
	const OjaRule &OjaDecision,
	const unsigned short &seed, 
	boost::random::mt19937 &rng,
	const unsigned short &DimIn, 
	const unsigned short &DimHidden, 
	const unsigned short &DimOut, 
	const unsigned short &TrainDim, 
	const unsigned short &TestDim, 
	const unsigned short &ValidationDim, 
	vector <float> learnCoeff1, 
	vector <float> learnCoeff2,
	const mat &TrainSet, 
	const mat &TestSet, 
	const mat &ValidationSet, 
	const mat &TargetTrSet,
	const mat &TargetTsSet,
	const mat &TargetVlSet,
	unsigned short numEpochs1, 
	unsigned short numEpochs2,
	const float &lr_dec, 
	const float &lr_inc, 
	const float &max_perf_inc, 
	const float &goal, 
	const float &momRate, 
	const float &actFcnFlatness, 
	const float &SatLimit, 
	const float &ThresholdSat,
	const float &AuxSigma,
	const float &AuxSigma2,//for sigma_corr
	const unsigned short &L,//for sigma_corr
	float &lrCoeff_sigma//for sigma_corr
	)
{
	

	#ifdef TimeFile
		time_t start,end; //count TIME by epoch
		time (&start); //TIME
	#endif
	//###end of defining a vector containning the learning coefficients to test the redutor###
	//enum TrainMethods Method = ITL_Method; //definir o mtodo aqui! ENTROPY, CAUCHY_SCHWARTZ, EUCLIDEAN_DISTANCE, UNI
	unsigned short DimLC1 = learnCoeff1.size(); 
	unsigned short DimLC2 = learnCoeff2.size();

	//string com o nome do metodo
	string MethodName = StringTrainMethods (W1_Method);
	
	//string para o ficheiro de results
	#ifdef ResultsFile
		string results = Label_data;
		results.append(MethodName);
		results.append("_Results.csv");
	#endif

	#ifdef MSE_Validation_file
		string ValidStr = Label_data;
		ValidStr.append(MethodName);
		ValidStr.append("_ValidationMSE.csv");
	#endif
	//ofstream para todos os ficheiros
	ofstream file;
	
	double dDimIn = static_cast<double>(DimIn); 
	double dDimHidden = static_cast<double>(DimHidden); 
	double dDimOut = static_cast<double>(DimOut); 

	double dTestDim = static_cast<double>(TestDim); 
	double dTrainDim = static_cast<double>(TrainDim); 
	
	//Same random initialization FOR each experiment without depending on choices of B1, B2, W1, W2
	rowvec Bias1Fixo(DimHidden);
	rowvec Bias2Fixo(DimOut);
	mat Weights1Fixo(DimIn, DimHidden);
	mat Weights2Fixo(DimHidden,DimOut);
	for (unsigned short hid=0; hid<DimHidden; hid++)
	{
		Bias1Fixo(hid) = aleat(rng);
	}
	for (unsigned short out=0; out<DimOut; out++)
	{
		Bias2Fixo(out) = aleat(rng);
	}
	for (unsigned short in=0; in<DimIn; in++)
	{
		for (unsigned short hid=0; hid<DimHidden; hid++)
		{
			Weights1Fixo(in,hid) = aleat(rng);
		}
	}
	for (unsigned short hid=0; hid<DimHidden; hid++)
	{
		for (unsigned short out=0; out<DimOut; out++) 
		{
			Weights2Fixo(hid,out) = aleat(rng);
		}
	}
	
	//##### Bias1 Inicialization #####
	switch (Init_B1_Method)
	{
		case RANDOM_B1: 
		{
			break;
		}
		
		case ZEROS_B1:
		{
			Bias1Fixo.zeros();
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	//##### End of Bias1 Inicialization #####

	//##### Weights1 Inicialization #####
	switch (Init_W1_Method)
	{
		case RANDOM_W1: 
		{
			break;
		}
		
		case PCA:
		{
			PCA_function(Weights1Fixo, TrainSet, DimIn, DimHidden);
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	//cout << "Weights1Fixo " << Weights1Fixo << endl;
	//##### End of Weights1 Inicialization #####

	//Initialization of vector Bias2
	switch (Init_B2_Method)
	{
		case RANDOM_B2: 
		{
			break;
		}
		
		case ZEROS_B2:
		{
			Bias2Fixo.zeros();
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	//END of Bias2 Initialization 


	//##### Weights2 Inicialization #####
	switch (Init_W2_Method)
	{
		case RANDOM_W2: 
		{
			break;
		}
		
		case TRANSPOSE:
		{
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	//##### End of Weights2 Inicialization #####


	//##### Hidden Matrix == FeedForward Step #####
	//cria matriz de HiddenTrain de (DimInensao padres X tamanhoOutput)
	mat HiddenTrain (TrainDim, DimHidden);
	transitionLayer (HiddenTrain, TrainDim, DimIn, DimHidden, TrainSet, Weights1Fixo, Bias1Fixo, actFcnFlatness, NeuronActFunct1);
	//##### End of Calculating Hidden Matrix #####


	//##### Sigma Calculation #####
	// Sigma Calculation By Silverman's Rule
	// Silverman, 1998 (First Edition 1986), page87
	double sig2 = sigmaFunction (DimHidden, TrainDim, HiddenTrain, sigmaMethod, AuxSigma);
	//cout << "sig2 " << sig2 << endl;
	//cout << "conta " << 1/(pow(sqrt(2.0*3.14159265359),12.0)*pow(2.0*sig2, 12.0)) << endl;
	//##### End of Sigma Calculation #####
	
	//as variveis improved so as que vo ser escolhidas entre todas as combinao de c1 e c2
	mat ImprovedW1 (DimIn, DimHidden);
	mat ImprovedW2 (DimHidden, DimOut);
	rowvec ImprovedB1 (DimHidden);
	rowvec ImprovedB2 (DimOut);
	double Improved_MSEtest = 1000.0;

	for (unsigned short v=0; v<DimLC1; v++)
	{
		mat Weights1 = Weights1Fixo;
		rowvec Bias1 = Bias1Fixo;
		float lrCoeff1 = learnCoeff1[v];   

		//string para o ficheiro de cfTest
		string cfTest = Label_data;
		cfTest.append(StringConstructor_function (learnCoeff1[v], MethodName, "CfTest_c1_"));


		//##### Maximum Learning Coeficient Calculation #####
		// CRITERIUM: lrCoeff < 1/ 2*lambdaMax
		float Max_lrCoeff1 = maxLrCoeffFunction (DimIn, TrainSet, learnCoeff1[v]);
		//##### End of Maximum Learning Coeficient Calculation  #####
		

		//######################### NN Training Inicialization #########################
		//######################### NN Training Inicialization #########################

		mat Vij_x(TrainDim,TrainDim); //for Cauchy Schwartz
		colvec Vi_x(TrainDim); //for Cauchy Schwartz
		double V_x=0.0; //for Cauchy Schwartz

		mat Vij_x_test (TestDim,TestDim); //for Cauchy Schwartz test
		colvec Vi_x_test (TestDim); //for Cauchy Schwartz test
		double V_x_test=0.0; //for Cauchy Schwartz test

		mat Cij(TrainDim, TrainDim); // for Euclidean Distance
		mat Cij_test (TestDim, TestDim); // for Euclidean Distance

		double maxORmin;

		switch (W1_Method)
		{
			case ENTROPY: 
			{
				maxORmin = 1.0;//Max Ent
				break;
			}
		
			case CAUCHY_SCHWARTZ:
			{
				maxORmin = 1.0;//Max MI

				Vij_x = Vij_x_CS_function (TrainDim, DimIn, TrainSet, sig2);
				Vi_x = sum(Vij_x, 1);
				V_x = accu(Vij_x);

				//for the testSet
				Vij_x_test = Vij_x_CS_function (TestDim, DimIn, TestSet, sig2);
				Vi_x_test = sum(Vij_x_test, 1);
				V_x_test = accu(Vij_x_test);
				break;
			}
			
			case EUCLIDEAN_DISTANCE: 
			{ 
				maxORmin = 1.0;//Max MI

				Cij = Cij_function (TrainDim, DimIn, sig2, TrainSet);
				Cij_test = Cij_function (TestDim, DimIn, sig2, TestSet);
				break;
			}

			case UNI:
			{
				maxORmin = 0.0;//MinMI Uni
				break;
			}
		
			default:
			{
				cout << "ERROR!!! ";
				break;
			}
		}
				

		double Prev_CostFn = 0.0;
		double CostFn = 100000.0; //grande para criterio adaptativo do lrCoeff nao cortar na 1a iteraao

		double Prev_CostFnTest = 0.0;
		double CostFnTest = 100000.0; //grande para criterio adaptativo do lrCoeff nao cortar na 1a iteraao
		double Best_CostFnTest = -5000.0; //armazena valor da melhor entropia(max); Caso MinMI  alterado no switch

		mat prev_Weights1(DimIn,DimHidden); prev_Weights1.zeros();
		rowvec prevBias1(DimHidden); prevBias1.zeros();

		mat Best_Weights1 (DimIn,DimHidden);
		rowvec Best_Bias1 (DimHidden);
			
		for (unsigned short Epochs1=0; Epochs1<numEpochs1; Epochs1++)
		{

			//cout << endl << endl << endl << endl << "epoca numero: " << Epochs1 << endl;
			prev_Weights1 = Weights1;
			prevBias1 = Bias1;

			switch (W1_Method)
			{
				case ENTROPY: 
				{
					Ent_Weights1Bias1_function (DimHidden, DimIn, TrainDim, HiddenTrain, TrainSet, actFcnFlatness, sig2, momRate, lrCoeff1, Weights1, prev_Weights1, Bias1, prevBias1, NeuronActFunct1);
					break;
				}
		
				case CAUCHY_SCHWARTZ:
				{
					CS_Weights1Bias1_function (DimHidden, DimIn, TrainDim, HiddenTrain, TrainSet, Vij_x, Vi_x, actFcnFlatness, sig2, momRate, lrCoeff1, Weights1, prev_Weights1, Bias1, prevBias1, NeuronActFunct1);
					break;
				}
			
				case EUCLIDEAN_DISTANCE: 
				{
					ED_Weights1Bias1_function (DimHidden, DimIn, TrainDim, HiddenTrain, TrainSet, Cij, actFcnFlatness, sig2, momRate, lrCoeff1, Weights1, prev_Weights1, Bias1, prevBias1, NeuronActFunct1);
					break;
				}

				case UNI:
				{
					Uni_Weights1Bias1_function (DimHidden, DimIn, TrainDim, HiddenTrain, TrainSet, Vij_x, Vi_x, actFcnFlatness, sig2, momRate, lrCoeff1, Weights1, prev_Weights1, Bias1, prevBias1, NeuronActFunct1);
					break;
				}
		
				default:
				{
					cout << "ERROR!!! ";
					break;
				}
			}


			switch (OjaDecision)
			{
				case OjaYES: 
				{
					OjaFunction (DimHidden, DimIn, Weights1, Bias1);
					break;
				}
				case OjaNO: 

				break;
				
				default: cout << "ERROR!!! ";
			}

			
			mat prev_HiddenTrain (TrainDim, DimHidden);
			prev_HiddenTrain= HiddenTrain;

			transitionLayer (HiddenTrain, TrainDim, DimIn, DimHidden, TrainSet, Weights1, Bias1, actFcnFlatness, NeuronActFunct1);
						
			Saturation_Process (TrainDim, DimIn, DimHidden, TrainSet, SatLimit, ThresholdSat, actFcnFlatness, NeuronActFunct1, HiddenTrain, Weights1, Bias1, rng);
								
			Prev_CostFn = CostFn;

			//##### Calculating Cost Function for this Hidden Matrix  #####
			switch (W1_Method)
			{
				case ENTROPY:  
				{
					CostFunction_Ent (CostFn, DimHidden, TrainDim, HiddenTrain, sig2);
					break;
				}
		
				case CAUCHY_SCHWARTZ:
				{
					CostFunction_CS (CostFn, DimHidden, TrainDim, HiddenTrain, sig2, Vij_x, Vi_x, V_x);
					break;
				}
			
				case EUCLIDEAN_DISTANCE: 
				{ 
					CostFunction_ED (CostFn, DimHidden, TrainDim, HiddenTrain, sig2, Cij);
					break;
				}

				case UNI:
				{
					CostFunction_Uni (CostFn, DimHidden, TrainDim, HiddenTrain, sig2, Vij_x, Vi_x, V_x);
					break;
				}
		
				default:
				{
					cout << "ERROR!!! ";
					break;
				}
			}
			//##### End of Calculating Cost Function for this Hidden Matrix  #####


			//##### Stop Criterium ##### //when maximizing the stop criteria is the number of epochs only
			/*if ( abs(CostFn) < goal)
			{
				Epochs1 = numEpochs1;
			}*/
			//##### End #####

			//##### Learning Coeff Adaptive #####
			learningAdaptBias(CostFn, Prev_CostFn, max_perf_inc, Bias1, prevBias1, maxORmin); // 1 indica maximizar a cost function
			learningAdaptMatrix(CostFn, Prev_CostFn, max_perf_inc, Weights1, prev_Weights1, maxORmin);
			learningAdaptMatrix(CostFn, Prev_CostFn, max_perf_inc, HiddenTrain, prev_HiddenTrain, maxORmin);
			learningAdaptCoeff (CostFn, Prev_CostFn, lrCoeff1, lr_inc, Max_lrCoeff1, max_perf_inc, lr_dec, maxORmin);	
			//##### End Learning Coeff Adaptative #####

			// ##### TEST COST FUNCTION ####
			//compress the TestSet 
			mat HiddenTest (TestDim, DimHidden);
			transitionLayer (HiddenTest, TestDim, DimIn, DimHidden, TestSet, Weights1, Bias1, actFcnFlatness, NeuronActFunct1);
					
			Prev_CostFnTest = CostFnTest;				
					
			//##### Calculating Cost Function for this Hidden Matrix  #####
			switch (W1_Method)
			{
				case ENTROPY: 
				{
					CostFunction_Ent (CostFnTest, DimHidden, TestDim, HiddenTest, sig2);
					break;
				}
		
				case CAUCHY_SCHWARTZ:
				{
					CostFunction_CS (CostFnTest, DimHidden, TestDim, HiddenTest, sig2, Vij_x_test, Vi_x_test, V_x_test);
					break;
				}
			
				case EUCLIDEAN_DISTANCE: 
				{ 
					CostFunction_ED (CostFnTest, DimHidden, TestDim, HiddenTest, sig2, Cij_test);
					break;
				}

				case UNI:
				{
					CostFunction_Uni (CostFnTest, DimHidden, TestDim, HiddenTest, sig2, Vij_x_test, Vi_x_test, V_x_test);
					break;
				}
		
				default:
				{
					cout << "ERROR!!! ";
					break;
				}
			}	

			//BEST condition
			switch (W1_Method)
			{
				case ENTROPY: 
				{
					if (CostFnTest>Best_CostFnTest)
					{
						Best_CostFnTest = CostFnTest;
						Best_Weights1 = Weights1;
						Best_Bias1 = Bias1;
					}
					break;
				}
		
				case CAUCHY_SCHWARTZ:
				{
					if (CostFnTest>Best_CostFnTest)
					{
						Best_CostFnTest = CostFnTest;
						Best_Weights1 = Weights1;
						Best_Bias1 = Bias1;
					}
					break;
				}
			
				case EUCLIDEAN_DISTANCE: 
				{ 
					if (CostFnTest>Best_CostFnTest)
					{
						Best_CostFnTest = CostFnTest;
						Best_Weights1 = Weights1;
						Best_Bias1 = Bias1;
					}
					break;
				}

				case UNI:
				{
					Best_CostFnTest = 5000.0;

					if (CostFnTest<Best_CostFnTest)
					{
						Best_CostFnTest = CostFnTest;
						Best_Weights1 = Weights1;
						Best_Bias1 = Bias1;
					}
					break;
				}
		
				default:
				{
					cout << "ERROR!!! ";
					break;
				}
			}	

			//cout << "Best_CostFnTest" << endl << Best_CostFnTest << endl;
			#ifdef CostFnTestFile
				file.open(cfTest, ios::app);
				file << "Seed;" << seed << ";Best_CostFnTest;"<< Best_CostFnTest << ";CostFnTest ;"<< CostFnTest<< endl;
				file.close();
			#endif
			// ##### TEST COST FUNCTION ####
		}
		
			
		for (unsigned short p=0; p<DimLC2; p++)
		{
			//##### constructing the file to include evolution of MSE over epochs for each lrCoeff2, considering all seeds #####
			float lrCoeff2 = learnCoeff2[p];
			
			//Ficheiro Resultados
			#ifdef ResultsFile;
			file.open(results, ios::app);
			file << "lrCoeff1 Inicial:;" << learnCoeff1[v] << ";lrCoeff1 Final:;" << lrCoeff1 << ";seed:;" << seed << ";CostFn:;" << CostFn<< ";";
			file.close();
			#endif
			//##### END OUTPUTs for each lrCoeff #####

		
			//####################### RProp (AUMENTADOR) ##########################
			//####################### inicio do AUMENTADOR ##########################
				
			//########################################################################################
				
			//###### maximum learnCoeff 2 value to be adopted #########
			float Max_lrCoeff2 = maxLrCoeffFunction (DimHidden, HiddenTrain, learnCoeff2[p]);			
			//###### maximum learnCoeff 2 value to be adopted #########
			//########################################################################################
			

			mat OutputsTrain(TrainDim, DimOut); 
			//mat delta_weights2(DimHidden,DimOut); //TO REMOVE
			
			Weights1 = Best_Weights1;
			Bias1 = Best_Bias1;

			mat Weights2;
			rowvec Bias2 = Bias2Fixo;
				
			//the second part weights may be initialized with the transposed final weights from first part
			switch (Init_W2_Method)
			{
				case RANDOM_W2: 
				{
					Weights2 = Weights2Fixo;
					break;
				}
		
				case TRANSPOSE:
				{
					Weights2 = trans(Weights1);
					break;
				}
			
				default:
				{
					cout << "ERROR!!! ";
					break;
				}
			}
							
				
			//### Calculo inicial das particulas aumentadas ###
			//faz o conjunto de particulas aumentado dado o conjunto de pesos actual, todas de uma vez
			transitionLayer (OutputsTrain, TrainDim, DimHidden, DimOut, HiddenTrain, Weights2, Bias2, actFcnFlatness, NeuronActFunct2);
			//### END Calculo inicial das particulas aumentadas ###
				
				
			unsigned short Epochs2 = 0;
			double MSE_train = 100000.0, Prev_MSE_train=0.0;

			double Best_MSE_train=100.0, Best_MSE_test=100.0; // nota Best_MSE_test=100.0 vem inicializado do main para ser lido no main
							
			mat Best_Weights2 (DimHidden,DimOut);
			rowvec Best_Bias2 (DimOut);
			float CorrespondenteMSEtrain = 0.0f;

			// calcular sig2_corr
			vec error_vector(L+1);
			double sig2_corr = sigmaFunction (DimOut, TrainDim, OutputsTrain, sigmaMethod, AuxSigma2);

			while (Epochs2 < numEpochs2)
			{
				rowvec prevBias2(DimOut);
				prevBias2 = Bias2;
		
				mat prev_Weights2(DimHidden,DimOut);
				prev_Weights2 = Weights2;
					
				
				switch (W2_Method)
				{
					case W2_PROP: 
					{
						W2_Prop_function (TrainDim, DimOut, DimHidden, TargetTrSet, OutputsTrain, HiddenTrain, lrCoeff2, actFcnFlatness, NeuronActFunct2, OjaDecision, Weights2, Bias2);
						break;
					}

					case CIM: 
					{
						if (Epochs2>0)
						{
							learningAdaptCoeff( error_vector(L),  error_vector(L-1), lrCoeff_sigma, lr_inc, 1.0f, max_perf_inc, lr_dec, 0 );//1.0f = Max_lrCoeff, 0= maxORmin (neste caso min)
							Adaptive_Sigma_function (sig2_corr, error_vector, lrCoeff_sigma, L, Epochs2);
						}
						W2_CIM_function(TrainDim, DimOut, DimHidden, TargetTrSet, OutputsTrain, HiddenTrain, sig2_corr, lrCoeff2, actFcnFlatness, NeuronActFunct2, OjaDecision, Weights2, Bias2);
						break;
					}

					case Corr:
					{
						// sigma update if epochs>0
						if (Epochs2>0)
						{
							learningAdaptCoeff( error_vector(L),  error_vector(L-1), lrCoeff_sigma, lr_inc, 1.0f, max_perf_inc, lr_dec, 0 );//1.0f = Max_lrCoeff, 0= maxORmin (neste caso min)
							Adaptive_Sigma_function (sig2_corr, error_vector, lrCoeff_sigma, L, Epochs2);
						}
						//weights update
						W2_Corr_function (TrainDim, DimOut, DimHidden, TargetTrSet, OutputsTrain, HiddenTrain, sig2_corr, lrCoeff2, actFcnFlatness, NeuronActFunct2, OjaDecision, Weights2, Bias2);
						break;
					}
		
					default: cout << "ERROR!!! ";
				}
				
				//###update das particulas aumentadas ###
				//faz o conjunto de particulas aumentado dado o conjunto de pesos actual, todas de uma vez
				mat prev_OutputsTrain(TrainDim,DimOut);
				prev_OutputsTrain = OutputsTrain;
					
				transitionLayer (OutputsTrain, TrainDim, DimHidden, DimOut, HiddenTrain, Weights2, Bias2, actFcnFlatness, NeuronActFunct2);
				//### END update das particulas aumentadas ###

				Saturation_Process (TrainDim, DimHidden, DimOut, HiddenTrain, SatLimit, ThresholdSat, actFcnFlatness, NeuronActFunct2, OutputsTrain, Weights2, Bias2, rng);	
									
				//############### Train data set MAE between target and output #################
				//Error matrix calculation
				mat ErrorMatrixTrain = ErrorMatrixFunction (TrainDim, DimOut, TargetTrSet /*new*/, OutputsTrain);

				//####### vec error_vector for sig_corr calculation
				double error_actual = mean(mean(ErrorMatrixTrain,0));
				for(unsigned short i=0; i<L; i++)
				{
					error_vector(i) = error_vector(i+1);
				}
				error_vector(L) = error_actual;


							
				#ifdef DetailsFile;
					// Ficheiros detailed
					string detail1 = Label_data;
					detail1.append(MethodName);
					detail1.append("_DETAIL_c1_");
					detail1.append(doubletoStringFunction (learnCoeff1[v]));
					string detail = StringConstructor_function (learnCoeff2[p], detail1, "c2_");
					file.open(detail, ios::app);
					file << "Seed; "<< seed <<";Epoca ;" << Epochs2 << "; C1 ;"<< learnCoeff1[v] << "; C2 ;"<< learnCoeff2[p];
					file.close();
				#endif
				//############## END calculating MAE between target and output  ################


				//########### Train data set MSE between target and output ################
				Prev_MSE_train = MSE_train;
				MSE_train = MSEfunction (TrainDim, DimOut, ErrorMatrixTrain);
				
				//BEST condition
				if (MSE_train<Best_MSE_train)
				{
					Best_MSE_train=MSE_train;
				} 

				// Ficheiros detailed
				#ifdef DetailsFile;
				file.open(detail, ios::app);
				file << "; MSE_Train ;" << MSE_train << "; Best_MSE_train ;" << Best_MSE_train << ";;;;";
				file.close();
				#endif;
				//####################### END Train data set MSE between target and output  ############################

				//##### Stop Criterium ##### 
				//if the MSE_treino is lower than the criterion goal, the epochs loop is closed
				if ( abs( MSE_train) < goal)
				{
					Epochs2 = numEpochs2;
				}
				//##### End #####

				//##### Learning Coeff Adaptive #####
				learningAdaptCoeff (MSE_train, Prev_MSE_train, lrCoeff2, lr_inc, Max_lrCoeff2, max_perf_inc, lr_dec, 0); // zero indica minimizar a cost function
				learningAdaptBias(MSE_train, Prev_MSE_train, max_perf_inc, Bias2, prevBias2, 0);
				learningAdaptMatrix(MSE_train, Prev_MSE_train, max_perf_inc, Weights2, prev_Weights2, 0);
				learningAdaptMatrix(MSE_train, Prev_MSE_train, max_perf_inc, OutputsTrain, prev_OutputsTrain, 0);
				//##### End Learning Coeff Adaptative #####

				// ############# Rede Treinada e Calculo do MAE para Conj Test##################
				mat OutputsTest = nnOutputsFunction (TestDim, DimIn, DimHidden, DimOut, Weights1, Weights2, Bias1, Bias2, actFcnFlatness, TestSet, NeuronActFunct1, NeuronActFunct2);

				//#######################calculo do MSE em relao ao target e ao output ############################
				mat ErrorMatrixTest = ErrorMatrixFunction (TestDim, DimOut, TargetTsSet/*new*/, OutputsTest);
					
				
				//#######################calculo do MSE_TESTE ############################
				double MSE_test = MSEfunction (TestDim, DimOut, ErrorMatrixTest);

				//BEST
				if (MSE_test<Best_MSE_test)
				{
					Best_MSE_test=MSE_test;
					Best_Weights2 = Weights2;
					Best_Bias2 = Bias2;
					CorrespondenteMSEtrain = static_cast<float>(MSE_train);
				} 
				
				// Ficheiros sada
				#ifdef DetailsFile;
				file.open(detail, ios::app);
				file << "; MSE_test ;"<< MSE_test << "; Best_MSE_test ;"<< Best_MSE_test << endl;
				file.close();
				#endif
				// END Ficheiros sada
				//#######################FIM calculo do MSE_test ############################
								
				
				Epochs2++;
			}

			//W2=Best_Weights2;
			//B2=Best_Bias2;
			
					

			//##### OUTPUTs for each lrCoeff #####
			#ifdef ResultsFile;
			file.open(results, ios::app);
			file << ";C2: ;"<<learnCoeff2[p] << ";;;Best_MSE_test;" << Best_MSE_test << ";" << endl;
			file.close();
			#endif
			//##### END OUTPUTs for each lrCoeff #####
			
			//funcao rede (W1,W2,B1,B2) retorna MSE_ValidationSet
			mat ValidationOutput = nnOutputsFunction (ValidationDim, DimIn, DimHidden, DimOut, Best_Weights1, Best_Weights2, Best_Bias1, Best_Bias2, actFcnFlatness, ValidationSet, NeuronActFunct1, NeuronActFunct2);
			double MSEValidation = Error_MSE_function (ValidationDim, DimIn, TargetVlSet /*new*/, ValidationOutput);
			
			#ifdef MSE_Validation_file;
			file.open(ValidStr, ios::app);
			file << "Seed:;" << seed <<";C1:;"<< learnCoeff1[v] <<";C2:; "<< learnCoeff2[p] << ";MSE_TrainCorrespondente;" << CorrespondenteMSEtrain << ";Best_MSE_Test:;"<< Best_MSE_test << ";MSE validationSet:;" << MSEValidation << endl;
			file.close();
			#endif
			
			//BEST assessment
			if (Best_MSE_test < Improved_MSEtest)
			{
				Improved_MSEtest=Best_MSE_test;
				ImprovedW2 = Best_Weights2;
				ImprovedB2 = Best_Bias2;
				ImprovedW1 = Best_Weights1;
				ImprovedB1 = Best_Bias1;
			}

				
		}//loop c2
	} //loop c1 
	
	// retorna os melhores pesos e bias de entre todos os c1 e c2 percorridos
	W1 = ImprovedW1; 
	B1 = ImprovedB1;
	W2 = ImprovedW2; 
	B2 = ImprovedB2;  
	//Best_MSE_test = Improved_MSEtest; // retorna o melhor MSE_test entre os varios c1 e c2 testados

	#ifdef TimeFile //TIME
	time (&end);
	string time = Label_data;
	time.append(MethodName);
	time.append ("_Time.csv");
	file.open (time, ios::app);
	file << "seed:;" << seed <<  ";#C1:;" <<DimLC1 << ";#C2:;" << DimLC2 <<  ";seconds;" <<  difftime (end,start) << ";DimIn;" << DimIn <<";DimHidden;" << DimHidden << ";" << endl; //TIME
	file.close();
	#endif
}