#pragma once

#include <iostream>
#include <math.h>
#include <stdio.h>

#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <stdlib.h>

#include <armadillo> //matrices

//boost libraries
#include <boost/random.hpp>

using namespace std;
using namespace arma;




rowvec Parameter1_function (
	const mat &TrainSet,  
	const NormalizationMethod &NormMethod, 
	const unsigned short &DimIn)
{
	rowvec Parameter1 (1);

	switch (NormMethod)
	{
		case ENTRANCE_MIN_MAX: 
		{
			//Parameter1 = MaxTrain
			Parameter1.resize(DimIn);
			Parameter1= max(TrainSet,0);
			break;
		}
		
		case GLOBAL_MIN_MAX:
		{
			//Parameter1 = MaxTrain
			Parameter1(0) = max(max(TrainSet,0));
			break;
		}

		case ENTRANCE_ZSCORE:
		{
			//Parameter1 = MeanTrain
			Parameter1.resize(DimIn);
			Parameter1= mean(TrainSet,0);
			break;
		}

		case GLOBAL_ZSCORE:
		{
			//Parameter1 = MeanTrain
			Parameter1(0) = mean(mean(TrainSet,0));
			break;
		}

		case NONE:
		{
			//Parameter1 is not necessary, so it is zeros
			Parameter1.zeros();
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	 
	 return (Parameter1);
}

rowvec Parameter2_function (
	const mat &TrainSet,  
	const NormalizationMethod &NormMethod, 
	const unsigned short &DimIn)
{
	rowvec Parameter2 (1);

	switch (NormMethod)
	{
		case ENTRANCE_MIN_MAX: 
		{
			//Parameter2 = MinTrain;
			Parameter2.resize(DimIn);
			Parameter2= min(TrainSet,0);
			break;
		}
		
		case GLOBAL_MIN_MAX:
		{
			//Parameter2 = MinTrain;
			Parameter2(0) = min(min(TrainSet,0));
			break;
		}

		case ENTRANCE_ZSCORE:
		{
			//Parameter2 = SDev;
			Parameter2.resize(DimIn);
			Parameter2= stddev(TrainSet,0);
			break;
		}

		case GLOBAL_ZSCORE:
		{
			unsigned short vecDim = TrainSet.size();
			unsigned short linhas = vecDim /DimIn;
			vec auxParameter2 (vecDim);
			for (unsigned short in=0 ; in<linhas; in++)
			{
				for (unsigned short j=0 ; j<DimIn; j++)
				{
					auxParameter2(in*DimIn+j) = TrainSet(in, j);
				}
			}
			//Parameter2 = SDev;
			Parameter2(0) = stddev(auxParameter2);
			break;
		}

		case NONE:
		{
			//Parameter2 is not necessary, so it is zeros
			Parameter2.zeros();
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	return (Parameter2);
}


mat EntranceNormalizationMinMax (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &notNormalized, 
	const rowvec &MaxTrain, 
	const rowvec &MinTrain)
{
	mat EntranceNormalization(linhas,colunas);

	for(unsigned short te=0; te<linhas; te++)
	{
		for (unsigned short in=0; in<colunas; in++)
		{
			EntranceNormalization(te,in) = (2*notNormalized(te,in) - MaxTrain(in) - MinTrain(in)) / (MaxTrain(in) - MinTrain(in));
		}
	}
	return EntranceNormalization;
}

mat EntranceNormalizationZScore (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &notNormalized, 
	const rowvec &MeanTrain, 
	const rowvec &SDevTrain)
{
	mat EntranceNormZScore(linhas,colunas);

	for(unsigned short te=0; te<linhas; te++)
	{
		for (unsigned short in=0; in<colunas; in++)
		{
			EntranceNormZScore(te,in) = ( notNormalized(te,in) - MeanTrain(in) ) / ( SDevTrain(in) );
		}
	}
	return EntranceNormZScore;
}


mat GlobalNormalizationMinMax (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &notNormalized, 
	const rowvec &MaxTrain, 
	const rowvec &MinTrain)
{
	mat GlobalNormalization(linhas,colunas);

	for(unsigned short k=0; k<linhas; k++)
	{
		for (unsigned short j=0; j<colunas; j++)
		{
			GlobalNormalization(k,j) = (2.0*notNormalized(k,j) - MaxTrain(0) - MinTrain(0)) / (MaxTrain(0) - MinTrain(0));//if data is yet normalized
		}
	}
	return (GlobalNormalization);
}


mat GlobalNormalizationZScore (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &notNormalized, 
	const rowvec &MeanTrain, 
	const rowvec &SDevTrain)
{
	mat GlobalNormZScore(linhas,colunas);
	for(unsigned short te=0; te<linhas; te++)
	{
		for (unsigned short in=0; in<colunas; in++)
		{
			GlobalNormZScore(te,in) = ( notNormalized(te,in) - MeanTrain(0) ) / ( SDevTrain(0) );
		}
	}
	return (GlobalNormZScore);
}

mat  Normalization_function (
	const rowvec &P1, 
	const rowvec &P2, 
	const mat &notNormalized, 
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const NormalizationMethod &NormMethod)
{
	mat NormMatrix (linhas, colunas);
	switch (NormMethod)
	{
		case ENTRANCE_MIN_MAX: 
		{
			//Parameter1 = MaxTrain; Parameter2 = MinTrain;
			NormMatrix = EntranceNormalizationMinMax (linhas, colunas, notNormalized, P1, P2);
			break;
		}
				
		case GLOBAL_MIN_MAX:
		{
			////Parameter1 = MaxTrain; Parameter2 = MinTrain;
			NormMatrix = GlobalNormalizationMinMax (linhas, colunas, notNormalized, P1, P2);
			break;
		}

		case ENTRANCE_ZSCORE:
		{
			//Parameter1 = MeanTrain; Parameter2 = SDevTrain;
			NormMatrix = EntranceNormalizationZScore (linhas, colunas, notNormalized, P1, P2);
			break;
		}

		case GLOBAL_ZSCORE:
		{
			//Parameter1 = MeanTrain; Parameter2 = SDevTrain;
			NormMatrix = GlobalNormalizationZScore (linhas, colunas, notNormalized, P1, P2);
			break;
		}

		case NONE:
		{
			//when the data is already normalized 
			NormMatrix = notNormalized;
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	return (NormMatrix);
}

mat NormalizationTrain_function (
	const rowvec &MaxTrain, 
	const rowvec &MinTrain, 
	const mat &notNormTrainSet, 
	const unsigned short &TrainDim, 
	const unsigned short &DimIn, 
	const NormalizationMethod &NormMethod)
{
	//note: this code applies a Batch approach, therefore sorting the order of the patterns does not induce any significant change on the final results 
	// we kept this randomization just because yes (in this case it does not induce any change on the final result - it would only be significant for incremental case)

	//// ##### Randomize TrainSet ######
	//rng.seed(3578);

	/*
	vec RandPosition(TrainDim);
	for (unsigned short i=0; i<TrainDim;i++)
	{
		RandPosition(i) = aux_aleat(rng);
	}
	uvec indices = sort_index(RandPosition);
	// ##### End Randomize TrainSet ######
	

	//randomize the matrix positions 
	mat sortNotNormTrainSet (TrainDim, DimIn);
	for(unsigned short tr=0; tr<TrainDim; tr++)
	{
		for (unsigned short in=0; in<DimIn; in++)
		{
			sortNotNormTrainSet(tr,in) = notNormTrainSet(indices(tr),in);
		}
	}
	//end randomize the matrix positions 
	*/
	mat normalized = Normalization_function (MaxTrain, MinTrain, notNormTrainSet /*sortNotNormTrainSet*/, TrainDim, DimIn, NormMethod);
	return (normalized);
}



mat ReadAndNorm_function (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const string &Filename, 
	const rowvec &MaxTrain, 
	const rowvec &MinTrain, 
	const NormalizationMethod &NormMethod)
{
	mat notNormalized = lerFicheiro	(linhas, colunas, Filename);
	mat normalized = Normalization_function (MaxTrain, MinTrain, notNormalized, linhas, colunas, NormMethod);
	return (normalized);
}



mat GlobalDESNormalizationMinMax (
	const unsigned short &linhas, 
	const unsigned short &colunas, 
	const mat &GlobalNormalization, 
	const rowvec &MaxTrain, 
	const rowvec &MinTrain)
{
	mat notNormalized(linhas,colunas);

	for(unsigned short k=0; k<linhas; k++)
	{
		for (unsigned short j=0; j<colunas; j++)
		{
			notNormalized(k,j) = ( (MaxTrain(0) - MinTrain(0)) * GlobalNormalization(k,j) + MaxTrain(0) + MinTrain(0) )/2.0;
		}
	}
	return (notNormalized);
}